from geosoup.common import Sublist, Handler, Opt, FTPHandler, Logger, Timer
from geosoup.raster import Raster, MultiRaster, GDAL_FIELD_DEF, GDAL_FIELD_DEF_INV
from geosoup.terrain import Terrain
from geosoup.exceptions import ObjectNotFound, UninitializedError, FieldError, \
    FileNotFound, TileNotFound, ImageProcessingError
from geosoup.vector import Vector, OGR_GEOM_DEF, OGR_TYPE_DEF, OGR_FIELD_DEF



from itertools import takewhile, repeat
from functools import wraps
from uuid import uuid4
from decimal import *
import numpy as np
import datetime
import logging
import socket
import fnmatch
import random
import psutil
import shutil
import ftplib
import time
import copy
import gzip
import sys
import os


__all__ = ['Sublist',
           'Handler',
           'FTPHandler',
           'Logger',
           'Timer',
           'Opt']

julian_days = {
                  1: (1, 31),
                  2: (32, 59),
                  3: (60, 90),
                  4: (91, 120),
                  5: (121, 151),
                  6: (152, 181),
                  7: (182, 212),
                  8: (213, 243),
                  9: (244, 273),
                  10: (274, 304),
                  11: (305, 334),
                  12: (335, 365)
                }


class Sublist(list):
    """
    Class to handle list operations
    """
    def __eq__(self,
               other):
        """
        Check for a = other
        return: List of indices
        """
        temp = list(i for i in range(0, len(self)) if self[i] == other)

        if len(temp) == 1:
            temp = temp[0]

        return temp

    def __gt__(self,
               other):
        """
        Check for a > other
        return: List of indices
        """
        temp = list(i for i in range(0, len(self)) if self[i] > other)

        if len(temp) == 1:
            temp = temp[0]

        return temp

    def __ge__(self,
               other):
        """
        Check for a >= other
        return: List of indices
        """
        temp = list(i for i in range(0, len(self)) if self[i] >= other)

        if len(temp) == 1:
            temp = temp[0]

        return temp

    def __lt__(self,
               other):
        """
        Check for a < other
        return: List of indices
        """
        temp = list(i for i in range(0, len(self)) if self[i] < other)

        if len(temp) == 1:
            temp = temp[0]

        return temp

    def __le__(self,
               other):
        """
        Check for a <= other
        return: List of indices
        """
        temp = list(i for i in range(0, len(self)) if self[i] <= other)

        if len(temp) == 1:
            temp = temp[0]

        return temp

    def __ne__(self,
               other):
        """
        Check for a != other
        return: List of indices
        """
        temp = list(i for i in range(0, len(self)) if self[i] != other)

        if len(temp) == 1:
            temp = temp[0]

        return temp

    def __add__(self,
                other):
        """
        Adding another list
        :param other: other list
        :return: list
        """
        return Sublist(list(self) + list(other))

    def __getitem__(self,
                    item):
        """
        Method to get item(s) from the list using list or number as index
        :param item: Number or list of numbers
        :return: List
        """

        try:
            if isinstance(item, list):
                return list(list(self)[i] for i in item)
            else:
                return list(self)[item]

        except (TypeError, KeyError):
            print("List index not a number or list of numbers")

    def flatten(self):
        """
        Flatten a Sublist object
        returns: 1D Sublist object
        """
        if len(self) == 0:
            return self
        if isinstance(self, list):
            return Sublist(self[0]).flatten() + Sublist(self[1:]).flatten()
        return self[:1] + Sublist(self[1:]).flatten()

    def range(self,
              llim,
              ulim,
              index=False):
        """
        Make a subset of the list by only including elements between ulim and llim.
        :param ulim: upper limit
        :param llim: lower limit
        :param index: If the function should return index of values that lie within the limits
        :return: List
        """
        if index:
            return list(i for i, x in enumerate(self) if llim <= x <= ulim)
        else:
            return list(x for x in self if llim <= x <= ulim)

    def add(self,
            elem):
        """
        Add two lists
        :param elem: Another list or element
        :return: list
        """
        if isinstance(elem, list):
            for val in elem:
                self.append(val)
            return self
        else:
            self.append(elem)
            return self

    def remove_by_loc(self,
                      elem):
        """
        Method to remove a Sublist element with index 'other'
        :param elem: Index or list of indices
        :return: Sublist
        """
        mask = np.zeros(len(self), dtype=bool) + True
        mask[set(list(elem))] = False
        return (np.array(self)[np.where(mask)]).tolist()

    def remove(self,
               elem):
        """
        Method to remove list item or sublist in a 1d int or float list
        :param elem: item or list
        :return: list
        """
        if type(elem).__name__ not in ('list', 'tuple', 'generator'):
            return (np.array(self)[~np.in1d(np.array(self), np.array(elem))]).tolist()

        else:
            elem_ = list(set(list(elem)))
            return (np.array(self)[~np.in1d(np.array(self), np.array(elem_))]).tolist()

    @staticmethod
    def list_size(query_list):
        """
        Find size of a list object even if it is a one element non-list
        :param query_list: List to be sized

        """

        if isinstance(query_list, list):
            return len(query_list)
        else:
            return 1

    def remove_by_percent(self,
                          percent):
        """
        Method to remove randomly selected elements by percentage in a list
        :param percent: Percentage (Range 0-100)
        :return: list
        """
        nelem = len(self)
        nelem_by_percent = int(round((float(nelem)*float(100 - percent))/float(100)))
        return random.sample(self, nelem_by_percent)

    def random_selection(self,
                         num=1,
                         systematic=False):

        """
        Method to select a smaller number of samples from the Samples object
        :param num: Number of samples to select
        :param systematic: If the numbers returned should be systematic
                           Used only for large number of samples
        :return: Samples object
        """

        arr = np.array(list(range(len(self))))

        if num >= len(self):
            return self

        elif systematic:
            diff_seq = self.custom_list(0, len(self)-1, step=int(float(len(self))/float(num)))
            temp = arr[np.array(diff_seq)]
            return self[temp]

        else:
            return self[np.random.choice(arr, size=num, replace=False).tolist()]

    def tuple_by_pairs(self):
        """
        Make a list of tuple pairs of consequetive list elements
        :return: List of tuples
        """

        return Sublist((self[i], self[i+1]) for i in range(len(self) - 1))

    @classmethod
    def custom_list(cls,
                    start,
                    end,
                    step=None):
        """
        List with custom first number but rest all follow same increment; integers only
        :param start: starting integer
        :param end: ending integer
        :param step: step integer
        :return: list of integers
        """

        # end of list
        end = end + step - (end % step)

        # initialize the list
        out = Sublist()

        # make iterator
        if step is not None:
            if start % step > 0:
                out.append(start)
                start = step * (start < step) + (start + step - (start % step)) * (start > step)
            iterator = range(start, end, step)
        else:
            step = 1
            iterator = range(start, step, end)

        # add the rest of the list to out
        for i in iterator:
            out.append(i)

        return out

    def sublistfinder(self,
                      pattern):
        """
        Find the location of sub-list in a python list, no repetitions in mylist or pattern
        :param pattern: shorter list
        :return: list of locations of elements in mylist ordered as in pattern
        """
        return Sublist(self.index(x) if x in self else -1 for x in pattern)

    @staticmethod
    def hist(var_list,
             nbins=20,
             minmax=None):

        """
        Method to generate a histogram from a list of numbers
        :param nbins: Number of bins to group the numbers in a list
        :param minmax: Tuple of minimum and maximum values to be used
                       as criteria of inclusion
        :returns: frequency list (list of elements in each bin), list of bin edges

        """

        if minmax is None:
            minmax = (min(var_list), max(var_list))

        bin_edges = Sublist.frange(minmax[0], minmax[1], div=nbins)
        freq_list = list(0 for _ in range(len(bin_edges) - 1))

        for elem in var_list:
            for ii in range(len(bin_edges) - 1):
                if bin_edges[ii] <= elem < bin_edges[ii + 1]:
                    freq_list[ii] += 1
                  
        return freq_list, bin_edges

    @staticmethod
    def hist_equalize(list_dicts,
                      num=None,
                      pctl=50,
                      nbins=20,
                      var=None,
                      minmax=None):

        """
        Method to clip the frequency bins at a threshold and randomly select the
        elements from the remaining elements inside the bin
        :param list_dicts: List of dictionaries with one of key as the specified variable
        :param var: Name of the key, variable of interest
        :param pctl: Bin clipping percentile, ignored if num is specified
        :param num: Bin clipping value
        :param nbins: Number of bins
        :param minmax: Tuple of minimum and maximum values to be used
                       as criteria of inclusion
        :returns: list of dictionaries with clipped bins based on num or pctl
        """

        if var is None:
            raise ValueError('Variable not specified')
        else:
            var_list = np.array(list(elem[var] for elem in list_dicts))

            if minmax is None:
                minmax = (np.min(var_list),
                          np.max(var_list))

            freq_list, bin_edges = np.histogram(var_list,
                                                bins=nbins,
                                                range=minmax)

            if num is None:
                pctl_val = Sublist.percentile(freq_list,
                                              pctl=pctl)
            else:
                pctl_val = num

            out_list = list()

            for ii in range(len(freq_list)):

                if ii == (len(freq_list) - 1):
                    indices = np.where((var_list >= bin_edges[ii]) &
                                       (var_list <= bin_edges[ii + 1]))[0]

                else:
                    indices = np.where((var_list >= bin_edges[ii]) &
                                       (var_list < bin_edges[ii + 1]))[0]

                if indices.shape[0] > pctl_val:
                    out_indices = np.random.choice(indices,
                                                   size=pctl_val,
                                                   replace=False)
                else:
                    out_indices = indices

                out_list += list(list_dicts[jj] for jj in out_indices.tolist())

            return out_list

    @staticmethod
    def calc_parabola_param(pt1, pt2, pt3):
        """
        define a parabola using three points
        :param pt1: First point (x,y)
        :param pt2: Second point (x,y)
        :param pt3: Third point (x,y)
        :return tuple of a, b, and c for parabola a(x^2) + b*x + c = 0
        """
        x1, y1 = pt1
        x2, y2 = pt2
        x3, y3 = pt3

        _m_ = (x1 - x2) * (x1 - x3) * (x2 - x3)
        a_param = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / _m_
        b_param = (x3 * x3 * (y1 - y2) + x2 * x2 * (y3 - y1) + x1 * x1 * (y2 - y3)) / _m_
        c_param = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / _m_

        return a_param, b_param, c_param

    @staticmethod
    def moving_average(arr,
                       n=3,
                       cascaded=False):
        """
        Method to smooth an array of numbers using moving array method
        :param arr: Input array
        :param n: Number of elements to consider for moving average. Must be an odd number (default: 3)
        :param cascaded: If the smoothing should be cascaded from the specified moving average to the lowest(1)
        :return: smoothed array
        """
        if type(arr) in (list, tuple, dict, set):
            arr_copy = np.array(list(copy.deepcopy(arr)))
        elif type(arr) == np.ndarray:
            arr_copy = arr.copy()
        else:
            raise ValueError("Input array type not understood")

        dtype = arr_copy.dtype

        if n < 3:
            raise ValueError('n cannot be less than 3')

        if cascaded:
            ker_list = list(ker_size for ker_size in reversed(range(n + 1)) if ker_size % 2 == 1)
            if ker_list[-1] == 0:
                ker_list = ker_list[:-1]
        else:
            if n % 2 == 0:
                raise ValueError('n cannot be an even number')
            ker_list = [n]

        for ker_size in ker_list:
            if ker_size > 1:
                tail = int((ker_size - 1) / 2)

                ret = np.cumsum(np.concatenate([arr_copy[0:tail],
                                                arr_copy,
                                                arr_copy[-tail:]]),
                                dtype=np.float32)

                ret[ker_size:] = ret[ker_size:] - ret[:-ker_size]

                arr_copy = ret[ker_size - 1:] / ker_size

        arr_copy = arr_copy.astype(dtype)

        if type(arr) in (list, tuple, dict, set):
            return arr_copy.tolist()
        else:
            return arr_copy

    def count_in_range(self,
                       llim,
                       ulim):
        """
        Find elements in a range
        :param ulim: upper limit
        :param llim: lower limit
        :return: count, list
        """
        return len([i for i in self if llim <= i < ulim])

    @classmethod
    def frange(cls,
               start,
               end,
               step=None,
               div=None):
        """
        To make list from float arguments
        :param start: start number
        :param end: end number
        :param step: step
        :param div: Number of divisions between start and end
        :return: list
        """
        if end > start:
            return_rev = False
        elif end < start:
            start, end = end, start
            return_rev = True
        else:
            raise ValueError("Start and end value are the same!")

        if div is not None:
            temp = Sublist(start + i * ((end-start)/float(div)) for i in range(0, div)) + [end]

        elif step is not None:
            if (end - start) % step == 0.0:
                temp = Sublist(start + i * step for i in range(0, int((end - start) / step))) + [end]
            else:
                temp = Sublist(start + i * step for i in range(0, int((end - start) / step) + 1)) + [end]

        else:
            raise ValueError("No step or division defined")

        if return_rev:
            return Sublist(reversed(temp))
        else:
            return temp

    def reverse(self):
        """reversed list"""
        return Sublist(reversed(self))

    def shuffle(self):
        """
        shuffle the list items
        :return: list
        """
        x_ = self.copy()
        np.random.shuffle(x_)
        return x_

    def copy(self):
        """returns copied instance"""
        return copy.deepcopy(self)

    @classmethod
    def column(cls,
               matrix,
               i):
        """
        Get column of a numpy matrix
        :param matrix: Numpy matrix
        :param i: index
        :return: List
        """
        mat = matrix[:, i].tolist()
        return Sublist(elem[0] for elem in mat)

    @classmethod
    def row(cls,
            matrix,
            i):
        """
        Get row of a numpy matrix
        :param matrix: Numpy matrix
        :param i: index
        :return: List
        """
        mat = matrix[i].tolist()
        return Sublist(mat[0])

    def mean(self):
        """
        calculate mean of an array
        :return: float
        """
        return float(sum(self)) / max(len(self), 1)

    def max(self):
        """
        Calculate max of the list
        :return:
        """
        if type(max(self)).__name__ == 'list':
            return max(self)[0]
        else:
            return max(self)

    def median(self):
        """
        calculate median of an array
        :return: float
        """
        return self.percentile(self, pctl=50)

    @staticmethod
    def percentile(arr,
                   pctl=95.0):
        """
        Method to output percentile in an array
        :param arr: input numpy array or iterable
        :param pctl: percentile value
        :return:
        """

        return np.percentile(arr, pctl, interpolation='nearest')

    @staticmethod
    def pctl_interval(arr, intvl=95.0):
        """
        Method to calculate the width of a percentile interval
        :param arr: input numpy array or iterable
        :param intvl: Interval to calculate (default: 95th percentile)
        :return: scalar
        """
        lower = Sublist.percentile(arr, (100.0 - intvl)/2.0)
        upper = Sublist.percentile(arr, 100.0 - (100.0 - intvl)/2.0)

        return np.abs((upper - lower)/2.0)

    @staticmethod
    def std_dev(arr):
        """
        Method to calculate standard deviation of an array
        :param arr: input array
        :return: standard deviation
        """
        return np.std(arr)

    @staticmethod
    def reduce(array,
               method='mean',
               axis=0):
        """
        Method to reduce a 2D or 3D array using a specific method
        :param array: Numpy array
        :param method: Method to use for reduction (options: mean, median, std_dev,
                                                        pctl_xx, min, max, default:mean
                                                             here xx is the percentile)
        :param axis: Axis to apply the reducer along
        :return: float
        """

        methods = {'mean': np.mean,
                   'median': np.median,
                   'std_dev': np.std,
                   'min': np.min,
                   'max': np.max,
                   'pctl': np.percentile}

        if 'pctl' in method:
            return methods['pctl'](array, int(method.split('_')[1]), axis=axis)
        elif method in ('mean', 'median', 'std_dev', 'min', 'max'):
            return methods[method](array, axis=axis)
        else:
            raise RuntimeError("Reducer {} is not implemented".format(method))


class Handler(object):
    """
    Class to handle file and folder operations
    """
    def __init__(self,
                 filename=None,
                 basename=None,
                 dirname=None):

        self.filename = filename

        try:
            self.basename = os.path.basename(filename)
        except:
            self.basename = basename

        try:
            self.dirname = os.path.dirname(filename)
        except:
            self.dirname = dirname

        self.sep = os.path.sep

    def __repr__(self):
        if self.filename is not None:
            return '<Handler for {}>'.format(self.filename)
        elif self.dirname is not None:
            return '<Handler for {}>'.format(self.dirname)
        elif self.basename is not None:
            return '<Handler for {}>'.format(self.basename)
        else:
            return '<Handler {s}____{s}>'.format(s=self.sep)

    def dir_create(self):
        """
        Create dir if it doesnt exist
        :param: directory to be created
        """
        if not os.path.exists(self.dirname):
            os.makedirs(self.dirname)

    def add_to_filename(self,
                        string=None,
                        timestamp=False):
        """
        Method to append a string to a filename, before the .xxx extension
        :param string: String to append
        :param timestamp: If timestamp (upto seconds) should be added to a filename, before the .xxx extension
                          This can be used in addition to the string
        :return:
        """

        if timestamp:
            timestamp = datetime.datetime.now().isoformat().replace('-', '').replace(':', '').split('.')[0]
        else:
            timestamp = ''

        if string is not None:
            string += timestamp
        else:
            string = timestamp

        file_name, file_ext = os.path.splitext(self.basename)

        file_name += string

        self.basename = file_name + file_ext

        if self.dirname is not None:
            return self.dirname + self.sep + self.basename
        else:
            return self.basename

    def file_rename_check(self):
        """
        Change the file name if already exists by incrementing trailing number
        :return filename
        """
        if not os.path.isfile(self.filename):
            self.dir_create()

        # if file exists then rename the input filename
        counter = 1
        while os.path.isfile(self.filename):
            components = self.basename.split('.')
            if len(components) < 2:
                self.filename = self.dirname + os.path.sep + self.basename + '_' + str(counter)
            else:
                self.filename = self.dirname + os.path.sep + ''.join(components[0:-1]) + \
                           '_(' + str(counter) + ').' + components[-1]
            counter = counter + 1
        return self.filename

    def file_remove_check(self):
        """
        Remove a file silently; if not able to remove, change the filename and move on
        :return filename
        """

        # if file does not exist, try to create dir
        if not os.path.isfile(self.filename):
            self.dir_create()

        # if file exists then try to delete or
        # get a filename that does not exist at that location
        counter = 1
        while os.path.isfile(self.filename):
            # print('File exists: ' + filename)
            # print('Deleting file...')
            try:
                os.remove(self.filename)
            except Exception as e:
                print(e)
                components = self.basename.split('.')
                if len(components) < 2:
                    self.filename = self.dirname + os.path.sep + self.basename + '_' + str(counter)
                else:
                    self.filename = self.dirname + os.path.sep + ''.join(components[0:-1]) + \
                               '_(' + str(counter) + ').' + components[-1]
                # print('Unable to delete, using: ' + filename)
                counter = counter + 1
        return self.filename

    def file_exists(self):
        """
        Check file existence
        :return: Bool
        """
        return os.path.isfile(self.filename)

    def dir_exists(self):
        """
        Check folder existence
        :return: Bool
        """
        return os.path.isdir(self.dirname)

    def file_delete(self):
        """
        Delete a file
        """
        if self.file_exists():
            os.remove(self.filename)

    def array2d_to_file(self,
                        array,
                        sep=",",
                        action="w"):
        """
        Method to write a numpy array to text file
        :param array: Numpy ndarray
        :param sep: Separator between array items for text output.
                    If "" (empty), a binary file is written, equivalent to file.write(a.tobytes())
        :param action: "w" for write, "a" for append, "r" for read
        :return: None
        """
        with open(self.filename, action) as f:
            for row in array:
                row_str = sep.join([str(elem) for elem in row]) + '\n'
                f.write(row_str)

    def copy_file(self,
                  other_dir,
                  replace=True,
                  move=False):
        """
        Copy a file from one location to another
        :param other_dir: Name of output directory
        :param replace: If the existing output file should be replaced?
        :param move: If the file should be moved instead of copy
        """
        if other_dir[-1] == '/':
            out_filename = other_dir + self.basename
        else:
            out_filename = other_dir + self.sep + self.basename

        if replace:
            out_filename = Handler(out_filename).file_remove_check()

        shutil.copy(self.filename,
                    out_filename)

        if move:
            self.file_delete()

    def copy_dir(self,
                 other_dir):

        """
        Method to copy a directory to another
        :param other_dir: Name of output directory
        """
        shutil.copytree(self.dirname,
                        other_dir)

    def file_lines(self,
                   nlines=True,
                   bufsize=102400):
        """
        Find number of lines or get text lines in a text or csv file
        :param nlines: If only the number of lines in a file should be returned
        :param bufsize: size of buffer to be read
        :return: list or number
        """
        with open(self.filename, 'r') as f:
            bufgen = takewhile(lambda x: x, (f.read(bufsize) for _ in repeat(None)))

            if nlines:
                val = sum(buf.count('\n') for buf in bufgen if buf)
            else:
                val = list()
                remaining = ''
                for buf in bufgen:
                    if buf:
                        temp_lines = (remaining + buf).split('\n')
                        if len(temp_lines) <= 1:
                            remaining += ''.join(temp_lines)
                        else:
                            val += temp_lines[:-1]
                            remaining = temp_lines[-1]
        return val

    def read_csv_as_array(self):
        """
        Read csv file as numpy array
        """
        csv_data = np.genfromtxt(self.filename,
                                 delimiter=',',
                                 names=True,
                                 encoding="utf-8",
                                 dtype=None)

        names = [csv_data.dtype.names[i] for i in range(0, len(csv_data.dtype.names))]
        return names, csv_data

    def read_text_by_line(self):
        """
        Read text file line by line and output a list of text lines
        """
        with open(self.filename) as f:
            content = f.readlines()
        return [x.strip() for x in content]

    def read_array_from_csv(self,
                            delim=",",
                            array_1d=False,
                            nodataval=None):
        """
        Read array from file and output a numpy array. There should be no header.
        :param delim: Delimiter (default ',')
        :param array_1d: Should the array be reshaped to 1-dimensional array? (default: False)
        :param nodataval: Data values to be removed from 1-dimensional array
                          (ignored if array_1d flag is false)
        :returns: Numpy array (2d or 1d)
        """
        with open(self.filename) as f:
            content = f.readlines()
            lines = [x.strip() for x in content]
            ncols = len(lines[0].split(delim))
            nrows = len(lines)
            arr = np.zeros((nrows, ncols))
            for i, line in enumerate(lines):
                arr[i:] = [float(elem.strip()) for elem in line.split(delim)]

        if array_1d:
            lenr, lenc = arr.shape
            arr_out = arr.reshape(lenr * lenc)
            if nodataval is not None:
                arr_out = arr_out[arr_out != nodataval]
            return arr_out
        else:
            return arr

    def extract_gz(self,
                   dirname=None,
                   add_ext=None):
        """
        Extract landsat file to a temp folder
        :param dirname: Folder to extract all files to
        :param add_ext: Extension to be added to extracted file
        """

        if add_ext is None:
            add_ext = ''
        else:
            add_ext = '.' + str(add_ext)

        # define a temp file with scene ID
        if dirname is not None:
            tempfile = dirname + self.sep + self.basename.split('.gz')[0] + add_ext
        else:
            tempfile = self.dirname + self.sep + self.basename.split('.gz')[0] + add_ext

        # remove if file already exists
        temp = Handler(tempfile)
        temp.file_remove_check()

        # extract infile to temp folder
        if self.filename.endswith(".gz"):
            Opt.cprint('Extracting {} to {}'.format(self.basename,
                                                    temp.basename))
            with gzip.open(self.filename, 'rb') as gf:
                file_content = gf.read()
                with open(tempfile, 'wb') as fw:
                    fw.write(file_content)

        else:  # not a tar.gz archive
            raise TypeError("Not a .gz archive")

    def get_size(self,
                 unit='kb',
                 precision=3,
                 as_long=True):
        """
        Function to get file size
        :param unit: Unit to get file size in (options: 'b', 'kb', 'mb', 'gb', 'tb', 'pb', 'bit')
        :param precision: Precision to report
        :param as_long: Should the output be returned as a long integer? This truncates any decimal value
        :return: long integer
        """
        size = os.path.getsize(self.filename)

        getcontext().prec = precision

        if unit not in ('b', 'kb', 'mb', 'gb', 'tb', 'pb', 'bit'):
            raise NotImplementedError('Unrecognized size unit')

        output = (unit == 'bit') * float(Decimal(size)*Decimal(2**3)) + \
                 (unit == 'b') * float(Decimal(size)) + \
                 (unit == 'kb') * float(Decimal(size)/Decimal(2**10)) + \
                 (unit == 'mb') * float(Decimal(size)/(Decimal(2**20))) + \
                 (unit == 'gb') * float(Decimal(size)/(Decimal(2**30))) + \
                 (unit == 'tb') * float(Decimal(size)/(Decimal(2**40))) + \
                 (unit == 'pb') * float(Decimal(size)/(Decimal(2**50)))

        output = round(output, precision)

        if as_long:
            output = int(output)

        return output

    def write_list_to_file(self,
                           input_list,
                           rownames=None,
                           colnames=None,
                           delim=", ",
                           append=False):
        """
        Function to write list to file
        with each list item as one line
        :param input_list: input text list
        :param rownames: list of row names strings
        :param colnames: list of column name strings
        :param delim: delimiter (default: ", ")
        :param append: if the lines should be appended to the file
        :return: write to file
        """
        if len(input_list) == 0:
            raise ValueError('Empty input list supplied')

        elif type(input_list[0]).__name__ in ('list', 'tuple'):
            if rownames is not None:
                if len(rownames) != len(input_list):
                    raise ValueError('Row name list does not have sufficient elements')
                else:
                    input_list = list([rownames[i]] + elem for i, elem in enumerate(input_list))

            if colnames is not None:
                if len(rownames) != len(input_list[0]):
                    raise ValueError('Column name list does not have sufficient elements')
                else:
                    input_list = [colnames] + input_list

            input_list = '\n'.join(list(delim.join(list(str(elem) for elem in line)) for line in input_list))

        elif type(input_list[0]).__name__ == 'str':
            input_list = '\n'.join(input_list)

        else:
            raise ValueError('Input list not in types: list of list, list of tuples, list of strings')

        # create dir path if it does not exist
        self.dir_create()

        # write to file
        if append:
            open_type = 'a'  # append
        else:
            open_type = 'w'  # write

        with open(self.filename, open_type) as fileptr:
            fileptr.write('{}\n'.format(input_list))

    def write_slurm_script(self,
                           job_name='pyscript',
                           time_in_mins=60,
                           cpus=1,
                           ntasks=1,
                           mem=2000,
                           array=False,
                           iterations=1,
                           **kwargs):
        """
        Write slurm script for the given parameters
        :param job_name: Name of the SLURM job
        :param time_in_mins: expected run time in minutes
        :param cpus: Number of CPUs requested
        :param ntasks: Number of tasks
        :param mem: Memory requested (in MB)
        :param array: (bool) If using job arrays
        :param iterations: Job array upper limit (e.g. 132 for 1-132 array)
        :param kwargs: key word arguments
        :return:
        """
        script_dict = {
            'bash': '#!/bin/bash',
            'job-name': '#SBATCH --job-name=' + job_name,
            'time': '#SBATCH --time=' + str(datetime.timedelta(minutes=time_in_mins)),
            'cpus': '#SBATCH --cpus-per-task=' + str(cpus),
            'ntasks': '#SBATCH --ntasks=' + str(ntasks),
            'mem': '#SBATCH --mem=' + str(mem),
            'partition': '#SBATCH --partition=all',
            'array_def': '#SBATCH --array=1-' + str(iterations),
            'array_out': '#SBATCH --output=/scratch/rm885/support/out/slurm-jobs/iter_%A_%a.out',
            'out': '#SBATCH --output=/scratch/rm885/support/out/slurm-jobs/slurm_%j.out',
            'date': 'date',
            'array_echo': 'echo "Job ID is"$SLURM_ARRAY_TASK_ID'
        }

        script_list = [script_dict['bash'],
                       script_dict['job-name'],
                       script_dict['time'],
                       script_dict['cpus'],
                       script_dict['mem'],
                       script_dict['ntasks']]

        if array:
            script_list += [script_dict['array_def'],
                            script_dict['array_out'],
                            script_dict['array_echo']]

        script_list.append(script_dict['date'])

        for key, value in kwargs.items():
            script_list.append(value)

        script_list.append(script_dict['date'])

        self.write_list_to_file(script_list)

    def write_numpy_array_to_file(self,
                                  np_array,
                                  colnames=None,
                                  rownames=None,
                                  delim=","):
        """
        Write numpy array to file
        :param np_array: Numpy 2d array to be written to file
        :param colnames: list of column name strings
        :param rownames: list of row name strings (should have the same length as np_array axis 0
        :param delim: Delimiter (default: ", ")
        """
        if colnames is not None:
            header = delim.join(colnames)
        else:
            header = None

        if len(np_array.shape) == 1:
            np_array = np_array[:, np.newaxis]

        if rownames is not None:
            np_array = np.hstack(np.array(rownames)[:, np.newaxis], np_array)
            if header is not None:
                header = delim + header

        np_list = []
        np_list += np_array.tolist()
        np_list = list(delim.join(list(str(elem) for elem in row_list))
                       for row_list in np_list)
        if header is not None:
            np_list = [header] + np_list

        with open(self.filename, 'w') as fileptr:
            for line in np_list:
                fileptr.write('{}\n'.format(str(line)))

    def read_from_csv(self,
                      return_dicts=False,
                      line_limit=None,
                      read_random=False,
                      percent_random=50.0,
                      verbose=False,
                      systematic=False):
        """
        Read csv as list of lists with header.
        Each row is a list and a sample point.
        :param return_dicts: If list of dictionaries should be returned (default: False)
        :param line_limit: Limits on the number of lines returned (default: None)
        :param read_random: If lines to be read should be randomly selected
        :param percent_random: What percentage of lines should be randomly selected (default: 50%)
        :param verbose: Display steps
        :param systematic: If the samples should be systematic instead of random (used only if read_random is True)
        :returns: dictionary of data
        """
        lines = list()
        index_list = list()

        n_lines = self.file_lines(nlines=True)
        if verbose:
            sys.stdout.write('Lines in file: {}\n'.format(str(n_lines)))

        counter = 0
        err_counter = 0
        colnames = None
        with open(self.filename, 'r') as f:

            if not read_random:
                if line_limit is None:
                    for line in f:
                        if counter == 0:
                            colnames = list(elem.strip() for elem in line.split(','))
                        elif len(line.split(',')) != len(colnames):
                            err_counter += 1
                        else:
                            lines.append(list(elem.strip() for elem in line.split(',')))
                        counter += 1

                else:
                    for line in f:
                        if counter == 0:
                            colnames = list(elem.strip() for elem in line.split(','))
                        elif len(line.split(',')) != len(colnames):
                            err_counter += 1
                        elif counter <= line_limit:
                            lines.append(list(elem.strip() for elem in line.split(',')))
                        else:
                            break
                        counter += 1

            else:
                if percent_random > 100.0:
                    percent_random = 100.0
                elif percent_random < 0.0:
                    percent_random = 0.0
                n_rand_lines = int((float(percent_random) / 100.0) * float(n_lines))

                if n_rand_lines == 0:
                    lines = []
                else:
                    if n_rand_lines == n_lines:
                        index_list = list(range(0, n_lines))
                    else:
                        index_list = sorted([0] + Sublist(range(1, n_lines)).random_selection(num=n_rand_lines,
                                                                                              systematic=systematic))
                    if line_limit is not None:
                        if n_rand_lines > line_limit:
                            index_list = index_list[:line_limit]

                    for il, line in enumerate(f):
                        if counter == 0:
                            colnames = list(elem.strip() for elem in line.split(','))
                            counter += 1
                        elif len(line.split(',')) != len(colnames):
                            err_counter += 1
                        elif il == index_list[counter]:
                            lines.append(list(elem.strip() for elem in line.split(',')))
                            counter += 1

        if verbose:
            Opt.cprint('Encountered {} errors in reading file: {}'.format(str(err_counter),
                                                                          self.basename))

        if len(lines) > 0:
            # convert to list
            if return_dicts:
                if verbose:
                    sys.stdout.write('Converting to Dictionaries...\n')
                return list(dict(zip(colnames, list(self.string_to_type(elem) for elem in feat)))
                            for feat in lines[1:])
            else:
                return {
                    'feature': list(list(self.string_to_type(elem) for elem in feat)
                                    for feat in lines[1:]),
                    'name': colnames,
                }
        else:
            return {
                'feature': [],
                'name': [],
            }

    @staticmethod
    def write_to_csv(list_of_dicts,
                     outfile=None,
                     append=False,
                     delimiter=','):

        """
        Method to write to a csv file
        :param list_of_dicts: list of dictionaries to be written
        :param outfile: output file name
        :param delimiter: Delimiter examples : ,  ; |
        :param append: If the output should be appended to an already exising file
        """

        if outfile is None:
            raise ValueError("No file name for writing")

        if type(list_of_dicts) not in (tuple, list):
            list_of_dicts = [list_of_dicts]

        if append:
            if os.path.isfile(outfile):
                with open(outfile, 'r') as f:
                    colname_string = f.readline().strip()
                if len(colname_string) > 0:
                    colnames = [elem.strip() for elem in colname_string.split(delimiter)]
                    lines = list(delimiter.join(list(str(elem_dict[colname]) for colname in colnames))
                                 for elem_dict in list_of_dicts)
                else:
                    colnames = list(list_of_dicts[0].keys())
                    lines = [delimiter.join(colnames)] + \
                        list(delimiter.join(list(str(elem_dict[colname]) for colname in colnames))
                             for elem_dict in list_of_dicts)
                with open(outfile, 'a') as f:
                    for line in lines:
                        f.write(line + '\n')
            else:
                colnames = list(list_of_dicts[0].keys())
                lines = [delimiter.join(colnames)] + \
                    list(delimiter.join(list(str(elem_dict[colname]) for colname in colnames))
                         for elem_dict in list_of_dicts)
                with open(outfile, 'w') as f:
                    for line in lines:
                        f.write(line + '\n')
        else:
            colnames = list(list_of_dicts[0].keys())
            lines = [delimiter.join(colnames)] + \
                list(delimiter.join(list(str(elem_dict[colname]) for colname in colnames))
                     for elem_dict in list_of_dicts)
            with open(outfile, 'w') as f:
                for line in lines:
                    f.write(line + '\n')

    def find_all(self,
                 pattern='*',
                 find_dirs=False,
                 verbose=False):
        """
        Find all the names that match pattern
        :param pattern: pattern to look for in the folder
        :param find_dirs: If folder names should be searched instead of files
        :param verbose: Should the files be displayed
        """
        result = []
        # search for a given pattern in a folder path
        if '*' in pattern:
            search_str = pattern
        else:
            search_str = '*' + pattern + '*'

        if self.dirname[-1] != self.sep:
            self.dirname += self.sep

        for root, dirs, files in os.walk(self.dirname):
            if find_dirs:
                for dirname in dirs:
                    if verbose:
                        Opt.cprint(os.path.join(root, dirname))

                    if fnmatch.fnmatch(dirname, search_str):
                        if str(root) in str(self.dirname) or str(self.dirname) in str(root):
                            result.append(os.path.join(root, dirname))

            else:
                for name in files:
                    if verbose:
                        Opt.cprint(os.path.join(root, name))

                    if fnmatch.fnmatch(name, search_str):
                        if str(root) in str(self.dirname) or str(self.dirname) in str(root):
                            result.append(os.path.join(root, name))
        if verbose:
            Opt.cprint('======================\nFound {} results for {} in {}:'.format(str(len(result)),
                                                                                       pattern,
                                                                                       self.dirname))
            for filename in result:
                Opt.cprint(filename)
            Opt.cprint('======================')
        return result  # list

    def find_files(self,
                   pattern='*'):
        return list(f_ for f_ in list(f for f in os.listdir(self.dirname)
                                      if os.path.isfile(os.path.join(self.dirname, f)))
                    if pattern in f_)

    @staticmethod
    def string_to_type(x):
        """
        Method to return name of the data type
        :param x: input item
        :return: string
        """
        if type(x).__name__ == 'str':
            x = x.strip()
            try:
                val = int(x)
            except ValueError:
                try:
                    val = float(x)
                except ValueError:
                    try:
                        val = str(x)
                    except:
                        val = None
            x = val
        return x


class Opt:
    """
    Class for optional miscellaneous methods
    """
    def __init__(self,
                 obj=None):
        """
        Instantiate Opt class
        """
        self.obj = obj  # mutable object
        self.hex = uuid4().get_hex()

    def __repr__(self):
        return "<Optional notice class>"

    @staticmethod
    def print_memory_usage():
        """
        Function to print memory usage of the python process
        :return print to console/output
        """
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss

        if 2**10 <= mem < 2**20:
            div = float(2**10)
            suff = ' KB'
        elif 2**20 <= mem < 2**30:
            div = float(2**20)
            suff = ' MB'
        elif 2**30 <= mem < 2**40:
            div = float(2**30)
            suff = ' GB'
        elif mem >= 2**40:
            div = float(2**40)
            suff = ' TB'
        else:
            div = 1.0
            suff = ' BYTES'

        print_str = 'MEMORY USAGE: {:{w}.{p}f}'.format(process.memory_info().rss / div, w=5, p=2) + suff

        Opt.cprint(print_str)

    @staticmethod
    def time_now():
        """
        Prints current time
        :return: print to console/output
        """
        Opt.cprint('CURRENT TIME: ' + str(datetime.datetime.now()))

    @staticmethod
    def cprint(text,
               newline='\n'):
        """
        Print to stdout and flush
        :param text: Text to print to stdout
        :param newline: Newline character, default '\n' but can be changed
                        to '' to concatenate consequetive outputs
        """
        sys.stdout.write(str(text) + newline)
        sys.stdout.flush()

    @staticmethod
    def __copy__(obj):
        """
        Make a copy of an object using python's copy module
        """
        return copy.deepcopy(obj)

    @staticmethod
    def temp_name():
        """
        Generate a temperory name based on current machine time with a .tmp extension
        """
        return 'T{}'.format(np.random.random().hex()) + \
               datetime.datetime.now().isoformat().replace(':', '')\
            .replace('.', '').replace('-', '').replace('T', '') + '.tmp'

    @staticmethod
    def lists_to_dict(value_list=None,
                      column_list=None):
        """
        Constructs a dictionary from two lists
        :param value_list: list of values in a dictionary
        :param column_list: list of dictionary keys
        :return: Dictionary
        """
        if column_list is not None:
            if len(value_list) != len(column_list):
                raise ValueError("Returned query does not match column list")
            else:
                return dict(zip(column_list,
                                value_list))

        else:
            return dict(zip(['column' + str(i + 1) for i in range(0, len(value_list))],
                            value_list))

    @staticmethod
    def dtype(input_data):

        """
        Method to output data type as string
        :param input_data: input data
        :return: string
        """
        try:
            val = type(input_data).__name__
        except (ValueError, SyntaxError):
            val = None

        return val

    @staticmethod
    def list_size(query_list):
        """
        Find size of a list object even if it is a one element non-list
        :param query_list:
        """

        if isinstance(query_list, list):
            return len(query_list)
        else:
            return 1

    @staticmethod
    def remove_strings(main_str, remove_str_list):
        """
        Remove a list of strings from a given string
        :param main_str: Input string
        :param remove_str_list: List of strings to be removed
        :return:
        """
        temp = ''
        for i in range(0, len(remove_str_list)):

            # split the strings
            temp_strings = main_str.split(remove_str_list[i])

            # remove the string and rejoin
            if remove_str_list[i] != '\n':
                temp = ''.join(list(string for string in temp_strings))
            else:
                temp = '\n'.join(list(string for string in temp_strings if string != ''))

            # reassign
            main_str = temp

        return temp

    @staticmethod
    def hostname():
        """
        Method to print machine hostname
        :return: String
        """
        return socket.gethostname()

    @staticmethod
    def copy_item(item):
        return copy.deepcopy(item)

    @staticmethod
    def growth_curve(n,
                     maxn=100,
                     mi=8):
        """
        Method to emulate growth curve. This is used to determine
        number of compute nodes as number of operations increase
        :param n: Number of operations required
        :param maxn: Maximum value of output
        :param mi: A coefficient (default: 8)
        :return: Integer
        """
        x = n
        mx = maxn
        y = mx * (lambda t: (1 + mx ** (-1.25 * t / mi - 0.5)) **
                            (-1.0))((float(2.0 * x - mi * mx)) / mx)
        return int(np.floor(y))

    @staticmethod
    def ymd_to_yday(ymd_string):
        """
        Method to return tuple of (julian day, year) from date of format YYYY-MM-DD
        :param ymd_string: String with date format YYYY-MM-DD
        :return: Tuple (julian day, year)
        """

        date = datetime.datetime.strptime(ymd_string, '%Y-%m-%d')
        tt = date.timetuple()
        return tt.tm_yday, date.year

    @staticmethod
    def is_leap(year):
        """
        Method to determine if a year is leap or not
        :param year: Year
        :return: Boolean (True or False)
        """
        return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)

    @staticmethod
    def get_date_ranges(start_date, end_date):
        """
        Method to convert date string of the format YYYY-MM-DD to list of julian day ranges per year
        :param start_date: date of format YYYY-MM-DD
        :param end_date: date of format YYYY-MM-DD
        :return: List of dictionaries of format {'start': 150, 'end': 210 , 'year': 2017 }
        """

        start_jday, start_year = Opt.ymd_to_yday(start_date)
        end_jday, end_year = Opt.ymd_to_yday(end_date)

        n_years = end_year - start_year + 1

        if n_years == 1:
            dates = [{'start': start_jday,
                      'end': end_jday,
                      'year': start_year}]

        elif n_years > 1:
            dates = list()

            start_dict = {'start': start_jday,
                          'end': 365,
                          'year': start_year}

            if Opt.is_leap(start_year):
                start_dict['end'] = 366

            dates.append(start_dict)

            for i in range(0, n_years - 2):

                temp_dict = {'start': 1,
                             'end': 365,
                             'year': start_year + i + 1}

                if Opt.is_leap(start_year + i + 1):
                    temp_dict['end'] = 366

                dates.append(temp_dict)

            end_dict = {'start': 1,
                        'end': end_jday,
                        'year': end_year}

            if Opt.is_leap(end_year):
                end_dict['end'] = 366

            dates.append(end_dict)

        else:
            raise ValueError("Invalid dates or bad count")

        return dates

    @staticmethod
    def custom_list(start,
                    end,
                    step=None):
        """
        List with custom first number but rest all follow same increment; integers only
        :param start: starting integer
        :param end: ending integer
        :param step: step integer
        :return: list of integers
        """

        # end of list
        end = end + step - (end % step)

        # initialize the list
        out = list()

        # make iterator
        if step is not None:
            if start % step > 0:
                out.append(start)
                if start < step:
                    start = step
                elif start > step:
                    start = start + step - (start % step)
            iterator = range(start, end, step)
        else:
            step = 1
            iterator = range(start, step, end)

        # add the rest of the list to out
        for i in iterator:
            out.append(i)

        return out


class FTPHandler(Handler):
    """Class to handle remote IO for ftp connection"""
    def __init__(self,
                 filename=None,
                 basename=None,
                 dirname=None,
                 ftpserv=None,
                 ftpuser=None,
                 ftppath=None,
                 ftppasswd=None,
                 ftpfilepath=None):

        super(FTPHandler, self).__init__(filename,
                                         basename,
                                         dirname)
        self.ftpserv = ftpserv
        self.ftppath = ftppath
        self.ftpuser = ftpuser
        self.ftppasswd = ftppasswd
        self.ftpfilepath = ftpfilepath
        self.conn = None

    def __repr__(self):
        if self.ftpserv is not None:
            return '<Handler for remote ftp {}>'.format(self.ftpserv)
        else:
            return '<Handler for remote ftp ____>'

    def connect(self):
        """Handle for ftp connections"""
        # define ftp
        ftp = ftplib.FTP(self.ftpserv)

        # login
        if self.ftpuser is not None:
            if self.ftppasswd is not None:
                ftp.login(user=self.ftpuser, passwd=self.ftppasswd)
            else:
                ftp.login(user=self.ftpuser)
        else:
            ftp.login()

        # get ftp connection object
        self.conn = ftp

    def disconnect(self):
        """close connection"""
        self.conn.close()

    def getfiles(self):
        """Copy all files from FTP that are in a list"""

        # connection
        ftp_conn = self.conn

        # get file(s) and write to disk
        if isinstance(self.ftpfilepath, list):
            filenamelist = [self.dirname + self.sep + Handler(ftpfile).basename
                            for ftpfile in self.ftpfilepath]

            for i in range(0, len(filenamelist)):
                self.filename = filenamelist[i]
                with open(self.filename, 'wb') as f:
                    try:
                        if ftp_conn.retrbinary("RETR {}".format(self.ftpfilepath[i]), f.write):
                            print('Copying file {} to {}'.format(self.basename,
                                                                 self.dirname))

                    except Exception as err:
                        Opt.cprint('File {} not found or already written\n Error: {}'.format(self.basename,
                                                                                             err))
        else:
            self.filename = self.dirname + Handler().sep + Handler(self.ftpfilepath).basename
            with open(self.filename, 'wb') as f:
                try:
                    if ftp_conn.retrbinary("RETR {}".format(self.ftpfilepath), f.write):
                        Opt.cprint('Copying file {} to {}'.format(Handler(self.ftpfilepath).basename,
                                                             self.dirname))
                except Exception:
                    Opt.cprint('File {} not found or already written'.format(self.basename))


class Logger(object):
    """
    Class for logging output.
    Can be instantialized to log outputs to a file or stream
    """

    def __init__(self,
                 name='default',
                 filename=None,
                 level='info',
                 stream=False):
        """
        Instantialize logger class object
        """

        self.logger = logging.getLogger(name)

        if level.lower() == 'debug':
            self.logger.setLevel(logging.DEBUG)
        elif level.lower() == 'info':
            self.logger.setLevel(logging.INFO)
        elif level.lower() == 'warn' or level.lower() == 'warning':
            self.logger.setLevel(logging.WARNING)
        else:
            self.logger.setLevel(logging.ERROR)

        # create a logging format
        formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(message)s')

        # file handler
        if filename is not None:
            file_handler = logging.FileHandler(filename,
                                               mode='w')
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

        # stream handler
        if stream:
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(formatter)
            self.logger.addHandler(stream_handler)

    def __repr__(self):
        return '<Logger class object>'

    def lprint(self,
               message_str=None,
               level='info'):
        """
        Method to print to logger
        """

        if level.lower() == 'info':
            self.logger.info(message_str)
        elif level.lower() == 'warn' or level.lower() == 'warning':
            self.logger.warning(message_str)
        elif level.lower() == 'error':
            self.logger.error(message_str)
        else:
            self.logger.debug(message_str)

    def close(self):
        """
        Close logger
        """
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)


class Timer:
    """
    Decorator class to compute time a function takes to execute

    example usage:

    @Timer.timing(True)
    def my_function(a, b, c):
        return a + b + c

    >Time it took to run my_function: 0.44 seconds

    """
    def __init__(self,
                 func):
        """
        Constructor
        :param func: Input function
        """
        self.func = func

    def __repr__(self):
        """
        Object representation
        :return: String
        """
        return '<Timer class for {}>'.format(self.func.__name__)

    def __call__(self, *args, **kwargs):
        """
        Function to compute timing for input function
        :return: Function and prints time taken
        """

        t1 = time.time()
        val = self.func(*args, **kwargs)
        t2 = time.time()

        # time to run
        t = self.display_time(t2 - t1)

        sys.stdout.write("Time it took to run {}: {}\n".format(self.func.__name__, t))
        return val

    @staticmethod
    def display_time(seconds,
                     precision=3):
        """
        method to display time in human readable format
        :param seconds: Number of seconds
        :param precision: Decimal precision
        :return: String
        """

        # define denominations
        intervals = [('weeks', 604800),
                     ('days', 86400),
                     ('hours', 3600),
                     ('minutes', 60),
                     ('seconds', 1)]

        # initialize list
        result = list()

        # coerce to float
        dtype = type(seconds).__name__
        if dtype != 'int' or dtype != 'long' or dtype != 'float':
            try:
                seconds = float(seconds)
            except (TypeError, ValueError, NameError):
                print("Type not coercible to Float")

        # break denominations
        for name, count in intervals:
            if name != 'seconds':
                value = seconds // count
                if value:
                    seconds -= value * count
                    if value == 1:
                        name = name.rstrip('s')
                    value = str(int(value))
                    result.append("{v} {n}".format(v=value,
                                                   n=name))
            else:
                value = "{:.{p}f}".format(seconds,
                                          p=precision)
                result.append("{v} {n}".format(v=value,
                                               n=name))

        # join output
        return ' '.join(result)

    @staticmethod
    def timing(func):
        """
        Function to compute timing for input function
        :return: Function and prints time taken
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            Wrapper for func
            :param args: Arguments
            :param kwargs: Key word arguments
            :return: Function func return
            """

            t1 = time.time()
            val = func(*args, **kwargs)
            t2 = time.time()

            # time to run
            t = Timer.display_time(t2 - t1)

            sys.stdout.write("Time it took to run {}: {}\n".format(func.__name__, t))
            return val

        return wrapper




import numpy as np
from geosoup.common import Handler, Opt, Sublist
from geosoup.exceptions import ImageProcessingError, ObjectNotFound
import warnings
import random
import time
import json
from osgeo import gdal, gdal_array, ogr, osr, gdalconst
np.set_printoptions(suppress=True)

# Tell GDAL to throw Python exceptions, and register all drivers
gdal.UseExceptions()
gdal.AllRegister()


__all__ = ['Raster',
           'MultiRaster',
           'GDAL_FIELD_DEF',
           'GDAL_FIELD_DEF_INV']


GDAL_FIELD_DEF = {
    'byte': gdal.GDT_Byte,
    'int': gdal.GDT_Int16,
    'long': gdal.GDT_Int32,
    'float': gdal.GDT_Float32,
    'double': gdal.GDT_Float64,
    'uint': gdal.GDT_UInt16,
    'ulong': gdal.GDT_UInt32,
}

GDAL_FIELD_DEF_INV = dict(list((v, k) for k, v in GDAL_FIELD_DEF.items()))

OGR_FIELD_DEF = {
    'int': ogr.OFTInteger,
    'long': ogr.OFTInteger,
    'float': ogr.OFTReal,
    'double': ogr.OFTReal,
    'str': ogr.OFTString,
    'bool': ogr.OFTInteger,
    'nonetype': ogr.OFSTNone,
    'none': ogr.OFSTNone
}

OGR_FIELD_DEF_INV = dict(list((v, k) for k, v in OGR_FIELD_DEF.items()))

OGR_TYPE_DEF = {
            'point': 1,
            'line': 2,
            'linestring': 2,
            'polygon': 3,
            'multipoint': 4,
            'multilinestring': 5,
            'multipolygon': 6,
            'geometry': 0,
            'no geometry': 100
}


class Raster(object):
    """
    Class to read and write rasters from/to files and numpy arrays
    """

    def __init__(self,
                 name,
                 array=None,
                 bnames=None,
                 metadict=None,
                 dtype=None,
                 shape=None,
                 transform=None,
                 crs_string=None,
                 nodatavalue=None,
                 datasource=None):

        self.array = array
        self.array_offsets = None  # (px, py, xoff, yoff)
        self.bnames = bnames
        self.datasource = datasource
        self.shape = shape
        self.transform = transform
        self.crs_string = crs_string
        self.name = name

        if type(dtype) == str:
            try:
                self.dtype = GDAL_FIELD_DEF[dtype]
            except Exception in (TypeError, ValueError, KeyError):
                try:
                    self.dtype = int(dtype)
                except Exception in (TypeError, ValueError, KeyError):
                    self.dtype = GDAL_FIELD_DEF['float']
        elif type(dtype) in (int, float):
            self.dtype = int(dtype)
        else:
            self.dtype = gdal.GDT_Float32

        self.metadict = metadict
        self.nodatavalue = nodatavalue
        self.tile_grid = list()
        self.ntiles = None
        self.bounds = None
        self.init = False
        self.stats = dict()

    def __repr__(self):

        if self.shape is not None:
            return "<raster {ras} of size {bands}x{rows}x{cols} ".format(ras=Handler(self.name).basename,
                                                                         bands=self.shape[0],
                                                                         rows=self.shape[1],
                                                                         cols=self.shape[2]) + \
                "datatype {dt} 'no-data' value {nd}>".format(dt=str(self.dtype),
                                                             nd=str(self.nodatavalue))
        else:
            return "<raster with path {ras}>".format(ras=self.name,)

    def initialize(self,
                   get_array=False,
                   band_order=None,
                   finite_only=True,
                   nan_replacement=0.0,
                   verbose=False):

        """
        Initialize a raster object from a file
        :param get_array: flag to include raster as 3 dimensional array (bool)
        :param band_order: band location list of integers (starting at 0; ignored if get_array is False)
        :param finite_only: flag to remove non-finite values from array (ignored if get_array is False)
        :param nan_replacement: replacement for all non-finite replacements
        :param verbose: If some steps should be displayed
        (ignored if finite_only, get_array is false)
        :return None
        """
        self.init = True

        raster_name = self.name

        if Handler(raster_name).file_exists() or 'vsimem' in self.name:
            fileptr = gdal.Open(raster_name)  # open file
            self.datasource = fileptr
            self.metadict = Raster.get_raster_metadict(file_name=raster_name)

        elif self.datasource is not None:
            fileptr = self.datasource
            self.metadict = Raster.get_raster_metadict(file_ptr=fileptr)

        else:
            raise ImageProcessingError('No datasource found')

        if band_order is None:
            band_order = list(range(fileptr.RasterCount))
        elif max(band_order) >= fileptr.RasterCount:
            raise ImageProcessingError('Band indices must be smaller than number of available bands')

        self.shape = [fileptr.RasterCount if not get_array else len(band_order),
                      fileptr.RasterYSize,
                      fileptr.RasterXSize]

        if self.dtype is not None:
            self.dtype = fileptr.GetRasterBand(1).DataType

        self.nodatavalue = fileptr.GetRasterBand(1).GetNoDataValue()

        # if get_array flag is true
        if get_array:
            if verbose:
                Opt.cprint('Reading bands: {}'.format(" ".join([str(b) for b in band_order])))

            nbands = len(band_order)

            # initialize array
            if self.array_offsets is None:
                array3d = np.zeros((nbands,
                                    self.shape[1],
                                    self.shape[2]),
                                   gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype))
            else:
                array3d = np.zeros((nbands,
                                    self.array_offsets[3],
                                    self.array_offsets[2]),
                                   gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype))

            # get band names
            names = list()

            # read array and store the band values and name in array
            for i, b in enumerate(band_order):
                bandname = fileptr.GetRasterBand(b + 1).GetDescription()

                if verbose:
                    Opt.cprint('Reading band {}'.format(bandname))
                names.append(bandname)

                if self.array_offsets is None:
                    array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray()
                else:
                    array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray(*self.array_offsets)

            # if flag for finite values is present
            if finite_only:
                if np.isnan(array3d).any() or np.isinf(array3d).any():
                    array3d[np.isnan(array3d)] = nan_replacement
                    array3d[np.isinf(array3d)] = nan_replacement
                    if verbose:
                        Opt.cprint("Non-finite values replaced with " + str(nan_replacement))
                else:
                    if verbose:
                        Opt.cprint("Non-finite values absent in file")

            # assign to empty class object
            self.array = array3d
            self.bnames = names

            self.transform = fileptr.GetGeoTransform()
            self.crs_string = fileptr.GetProjection()

        else:
            self.bnames = list(fileptr.GetRasterBand(i + 1).GetDescription()
                               for i in range(self.shape[0]))
            self.transform = fileptr.GetGeoTransform()
            self.crs_string = fileptr.GetProjection()

        self.bounds = self.get_bounds()

    def write_to_file(self,
                      outfile=None,
                      driver='GTiff',
                      add_overview=False,
                      resampling='nearest',
                      overviews=None,
                      verbose=False,
                      **kwargs):
        """
        Write raster to file, given all the properties
        :param self: Raster object
        :param driver: raster driver
        :param outfile: Name of output file
        :param add_overview: If an external overview should be added to the file (useful for display)
        :param resampling: resampling type for overview (nearest, cubic, average, mode, etc.)
        :param overviews: list of overviews to compute( default: [2, 4, 8, 16, 32, 64, 128, 256])
        :param verbose: If the steps should be displayed
        :param kwargs: keyword arguments for creation options
                       example:
                       geotiff creation options can be used as follows:
                         compress='lzw',
                         bigtiff='yes
        """
        creation_options = []
        if len(kwargs) > 0:
            for key, value in kwargs.items():
                creation_options.append('{}={}'.format(key.upper(),
                                                       value.upper()))
        if outfile is None:

            if driver == 'MEM':
                outfile = 'tmp'
            else:
                outfile = self.name
                outfile = Handler(filename=outfile).file_remove_check()

        if verbose:
            Opt.cprint('\nWriting {}\n'.format(outfile))

        gtiffdriver = gdal.GetDriverByName(driver)
        fileptr = gtiffdriver.Create(outfile, self.shape[2], self.shape[1],
                                     self.shape[0], self.dtype, creation_options)
        nbands = self.shape[0]
        fileptr.SetGeoTransform(self.transform)
        fileptr.SetProjection(self.crs_string)

        if len(self.bnames) > 0:
            for i, bname in enumerate(self.bnames):
                if len(bname) == 0:
                    self.bnames[i] = 'band_{}'.format(str(i + 1))
        else:
            for i in range(self.shape[0]):
                self.bnames[i] = 'band_{}'.format(str(i + 1))

        if self.array is None:
            self.read_array()

        for i in range(0, nbands):
            fileptr.GetRasterBand(i + 1).WriteArray(self.array[i, :, :], 0, 0)
            fileptr.GetRasterBand(i + 1).SetDescription(self.bnames[i])

            if self.nodatavalue is not None:
                fileptr.GetRasterBand(i + 1).SetNoDataValue(self.nodatavalue)
            if verbose:
                Opt.cprint('Writing band: ' + self.bnames[i])

        if driver == 'MEM':
            if verbose:
                Opt.cprint('File written to memory!')
            return fileptr

        else:
            fileptr.FlushCache()
            fileptr = None

            if verbose:
                Opt.cprint('File written to disk!')

            if add_overview:
                if verbose:
                    Opt.cprint('\nWriting overview')

                out_ras = Raster(outfile)
                out_ras.add_overviews(resampling,
                                      overviews,
                                      **kwargs)
                out_ras = None

                if verbose:
                    Opt.cprint('Overview written to disk!')

    def add_overviews(self,
                      resampling='nearest',
                      overviews=None,
                      **kwargs):
        """
        Method to create raster overviews
        :param resampling:
        :param overviews:
        :param kwargs:
        :return:
        """

        fileptr = gdal.Open(self.name, 0)

        if overviews is None:
            overviews = [2, 4, 8, 16, 32, 64, 128, 256]

        if type(overviews) not in (list, tuple):
            if type(overviews) in (str, float):
                try:
                    overviews = [int(overviews)]
                except Exception as e:
                    Opt.cprint(e.args[0])
            elif type(overviews) == int:
                overviews = [overviews]
            else:
                raise ValueError('Unsupported data type for overviews list')
        else:
            if any(list(type(elem) != int for elem in overviews)):
                overviews_ = list()
                for elem in overviews:
                    try:
                        overviews_.append(int(elem))
                    except Exception as e:
                        Opt.cprint('Conversion error: {} -for- {}'.format(e.args[1], elem))

                overviews = overviews_

        for k, v in kwargs.items():
            gdal.SetConfigOption('{}_OVERVIEW'.format(k.upper()), v.upper())

        fileptr.BuildOverviews(resampling.upper(), overviews)
        fileptr = None

    def read_array(self,
                   offsets=None,
                   band_order=None,
                   n_tries=5,
                   nodatavalue=-9999):
        """
        Method to read raster array with offsets and a specific band order
        :param offsets: tuple or list - (xoffset, yoffset, xcount, ycount)
        :param band_order: order of bands to read
        :param n_tries: number of tries after read error
        :param nodatavalue: No data value
        """

        if not self.init:
            self.initialize()

        fileptr = self.datasource

        nbands, nrows, ncols = self.shape

        if offsets is None:
            self.array_offsets = (0, 0, ncols, nrows)
        else:
            self.array_offsets = offsets

        array3d = np.zeros((nbands,
                            self.array_offsets[3],
                            self.array_offsets[2]),
                           gdal_array.GDALTypeCodeToNumericTypeCode(fileptr.GetRasterBand(1).DataType))

        # read array and store the band values and name in array
        if band_order is not None:
            for b in band_order:
                self.bnames.append(self.datasource.GetRasterBand(b + 1).GetDescription())
        else:
            band_order = list(range(nbands))

        # read array and store the band values and name in array
        for i, b in enumerate(band_order):
            read_success = True
            for i_try in range(n_tries):
                try:
                    if self.array_offsets is None:
                        array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray()
                    else:
                        array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray(*self.array_offsets,
                                                                                    resample_alg=
                                                                                    gdalconst.GRA_NearestNeighbour)
                    break
                except RuntimeError as e:
                    read_success = False
                    Opt.cprint(e)
                    Opt.cprint('Retrying failed read... attempt {}'.format(str(i_try + 1)))
                    time.sleep(random.random() + 1.0)

            if not read_success:
                Opt.cprint('Retrying failed with {} attempts...using no-data value'.format(str(n_tries)))

                if self.nodatavalue is not None:
                    array3d[i, :, :] = self.nodatavalue
                else:
                    array3d[i, :, :] = nodatavalue

        if (self.shape[0] == 1) and (len(array3d.shape) > 2):
            self.array = array3d.reshape([self.array_offsets[3],
                                          self.array_offsets[2]])
        else:
            self.array = array3d

    def nodata_mask(self,
                    use_band=0,
                    outfile=None,
                    in_memory=False,
                    add_overviews=False,
                    out_format='GTiff',
                    data_type=gdal.GDT_Byte):
        """
        Method to write a mask for the input raster using nodata value as 0 and valid pixels as 1
        :param use_band: Band index to use for making the mask (Default: 0)
        :param outfile: Output file path
        :param in_memory: If the raster should be stored in memory (Default: False)
        :param add_overviews: If overviews should be added,
                              ignored if in_memory is True (Default: False)
        :param out_format: output file format,
                           ignored if in_memory is True (Default: 'GTiff')
        :param data_type: Output data type (default: Byte)
        """

        self.initialize(get_array=True,
                        band_order=[use_band])

        if not in_memory:
            if outfile is None:
                outfile = Handler(self.name).add_to_filename('_mask')

        temp_arr = self.array[0, :, :]
        temp_arr[np.where(temp_arr != self.nodatavalue)] = 1
        temp_arr[np.where(temp_arr == self.nodatavalue)] = 0

        temp_ras = Raster('temp_ras')
        temp_ras.nodatavalue = self.nodatavalue
        temp_ras.array = temp_arr
        temp_ras.shape = [1, self.shape[1], self.shape[2]]
        temp_ras.transform = self.transform
        temp_ras.crs_string = self.crs_string
        temp_ras.dtype = data_type
        temp_ras.bnames = ['mask']

        if in_memory:
            temp_ras.datasource = temp_ras.write_to_file(driver='MEM')
            return temp_ras

        else:
            temp_ras.write_to_file(outfile=outfile,
                                   driver=out_format,
                                   add_overviews=add_overviews)

    def set_nodataval(self,
                      in_nodataval=255,
                      out_nodataval=0,
                      outfile=None,
                      in_array=True,
                      **kwargs):
        """
        replace no data value in raster and write to tiff file
        :param in_nodataval: no data value in input raster
        :param out_nodataval: no data value in output raster
        :param in_array: if the no data value should be changed in raster array
        :param outfile: output file name
        """
        if in_array:
            if not self.init:
                self.initialize(get_array=True,
                                **kwargs)
            self.array[np.where(self.array == in_nodataval)] = out_nodataval

        self.nodatavalue = out_nodataval

        if outfile is not None:
            self.write_to_file(outfile)

    @property
    def chk_for_empty_tiles(self):
        """
        check the tile for empty bands, return true if one exists
        :return: bool
        """
        if Handler(self.name).file_exists():
            fileptr = gdal.Open(self.name)

            filearr = fileptr.ReadAsArray()
            nb, ns, nl = filearr.shape

            truth_about_empty_bands = [np.isfinite(filearr[i, :, :]).any() for i in range(0, nb)]

            fileptr = None

            return any([not x for x in truth_about_empty_bands])
        else:
            raise ObjectNotFound("File does not exist.")

    @staticmethod
    def get_raster_metadict(file_name=None,
                            file_ptr=None):
        """
        Function to get all the spatial metadata associated with a geotiff raster
        :param file_name: Name of the raster file (includes full path)
        :param file_ptr: Gdal file pointer
        :return: Dictionary of raster metadata
        """
        if file_name is not None:
            if Handler(file_name).file_exists():
                # open raster
                img_pointer = gdal.Open(file_name)
            else:
                raise ObjectNotFound("File does not exist.")

        elif file_ptr is not None:
            img_pointer = file_ptr

        else:
            raise ObjectNotFound("File or pointer not found")

        # get tiepoint, pixel size, pixel rotation
        geometadata = img_pointer.GetGeoTransform()

        # make dictionary of all the metadata
        meta_dict = {'ulx': geometadata[0],
                     'uly': geometadata[3],
                     'xpixel': abs(geometadata[1]),
                     'ypixel': abs(geometadata[5]),
                     'rotationx': geometadata[2],
                     'rotationy': geometadata[4],
                     'datatype': img_pointer.GetRasterBand(1).DataType,
                     'columns': img_pointer.RasterXSize,  # columns from raster pointer
                     'rows': img_pointer.RasterYSize,  # rows from raster pointer
                     'bands': img_pointer.RasterCount,  # bands from raster pointer
                     'projection': img_pointer.GetProjection(),  # projection information from pointer
                     'name': Handler(file_name).basename}  # file basename

        # remove pointer
        img_pointer = None

        return meta_dict

    def get_stats(self,
                  verbose=False,
                  approx=False):

        """
        Method to compute statistics of the raster object, and store as raster property
        :param verbose: If the statistics should be printed to console
        :param approx: If approx statistics should be calculated instead to gain speed
        :return: None
        """

        for ib in range(self.shape[0]):
            band = self.datasource.GetRasterBand(ib+1)
            band.ComputeStatistics(approx)
            band_stats = dict(zip(['min', 'max', 'mean', 'stddev'], band.GetStatistics(int(approx), 0)))

            if verbose:
                Opt.cprint('Band {} : {}'.format(self.bnames[ib],
                                                 str(band_stats)))

            self.stats[self.bnames[ib]] = band_stats

    def get_bands(self,
                  bands,
                  outfile=None,
                  return_vrt=False,
                  return_raster=True):
        """
        Method to return a raster or vrt with the specified bands or write a file if specified
        :param bands: List of band indices or names, cannot be mixed (index starts at 0)
        :param outfile: Output filename
        :param return_vrt: If a vrt object should be returned
        :param return_raster: If a Raster object should be returned
        """

        if not (isinstance(bands, list) or isinstance(bands, tuple)):
            bands = [bands]

        if type(bands[0]) == str:
            bands = list(self.bnames.index(elem) + 1 for elem in bands)
        else:
            bands = [elem + 1 for elem in bands]

        vrt_opts = gdal.BuildVRTOptions(bandList=bands)

        vrt = gdal.BuildVRT('/vsimem/{}.vrt'.format(Opt.temp_name()),
                            self.datasource,
                            options=vrt_opts)

        if outfile is not None:
            gdal.Translate(outfile,
                           vrt)
            vrt = None
            return
        elif return_raster:
            ras = Raster('bandSubsetRaster', datasource=vrt)
            ras.initialize()
            return ras
        elif return_vrt:
            return vrt
        else:
            raise RuntimeError("No further valid path")

    def rescale(self,
                out_type=None,
                rescale=False,
                scale_vals=None,
                scale_approx=False):

        """
        Method to rescale raster or to change the raster data type
        :param out_type: Out data type.
                         Valid options: byte, int, long, float, double
                                        int8, int16, int32, int64,
                                        float32, float64,
                                        uint8, uint16, uint32
        :param rescale: If the image should be rescaled
        :param scale_vals: Tuple of min and max values for rescaling
        :param scale_approx: If approx stats should be calculated
        :return: None
        """

        if out_type is None:
            out_type = gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype)
        else:
            common_types = {'byte': 'uint8',
                            'int': 'int16',
                            'float': 'float32',
                            'double': 'float64',
                            'long': 'int32'}

            if out_type in common_types:
                out_type = common_types[out_type]

        if gdal_array.NumericTypeCodeToGDALTypeCode(np.dtype(out_type)) != self.dtype:

            if rescale:
                if scale_vals is not None:

                    if self.stats is None:
                        self.get_stats(approx=scale_approx)
                    elif len(self.stats) < self.shape[0]:
                        self.get_stats(approx=scale_approx)

                    out_arr = np.zeros(self.shape, dtype=out_type)

                    for band_indx in range(self.shape[0]):
                        min_val, max_val = self.stats[band_indx]['min'], self.stats[band_indx]['max']

                        mdiff = float(scale_vals[1] - scale_vals[0]) / float(max_val - min_val)

                        out_arr[band_indx, :, :] = ((self.array[band_indx, :, :].astype(np.float64) - min_val) * mdiff +
                                                    scale_vals[0]).astype(out_type)

                    self.array = out_arr
                else:
                    self.array = self.array.astype(out_type)
            else:
                self.array = self.array.astype(out_type)

            self.dtype = gdal_array.NumericTypeCodeToGDALTypeCode(self.array.dtype)

            if self.nodatavalue is not None:
                self.nodatavalue = np.array(self.nodatavalue).astype(out_type).item()

    def make_polygon_geojson_feature(self):
        """
        Make a feature geojson for the raster using its metaDict data
        """

        meta_dict = self.metadict

        if meta_dict is not None:
            return {"type": "Feature",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [[
                             [meta_dict['ulx'], meta_dict['uly']],
                             [meta_dict['ulx'], meta_dict['uly'] - (meta_dict['ypixel'] * (meta_dict['rows'] + 1))],
                             [meta_dict['ulx'] + (meta_dict['xpixel'] * (meta_dict['columns'] + 1)),
                              meta_dict['uly'] - (meta_dict['ypixel'] * (meta_dict['rows'] + 1))],
                             [meta_dict['ulx'] + (meta_dict['xpixel'] * (meta_dict['columns'] + 1)), meta_dict['uly']],
                             [meta_dict['ulx'], meta_dict['uly']]
                             ]]
                        },
                    "properties": {
                        "name": meta_dict['name'].split('.')[0]
                        },
                    }
        else:
            raise AttributeError("Metadata dictionary does not exist.")

    @staticmethod
    def get_coords(xy_list,
                   pixel_size,
                   tie_point,
                   pixel_center=True):
        """
        Method to convert pixel locations to image coords
        :param xy_list: List of tuples [(x1,y1), (x2,y2)....]
        :param pixel_size: tuple of x and y pixel size. The signs of the pixel sizes (+/-) are as in GeoTransform
        :param tie_point: tuple of x an y coordinates of tie point for the xy list
        :param pixel_center: If the center of the pixels should be returned instead of the top corners (default: True)
        :return: List of coordinates in tie point coordinate system
        """

        if type(xy_list) != list:
            xy_list = [xy_list]

        add_const = (float(pixel_size[0])/2.0) * pixel_center, \
                    (float(pixel_size[1])/2.0) * pixel_center

        return list((float(xy[0]) * float(pixel_size[0]) + tie_point[0] + add_const[0],
                     float(xy[1]) * float(pixel_size[1]) + tie_point[1] + add_const[1])
                    for xy in xy_list)

    @staticmethod
    def get_locations(coords_list,
                      pixel_size,
                      tie_point):
        """
        Method to convert global coordinates to image pixel locations
        :param coords_list: Lit of coordinates in image CRS [(x1,y1), (x2,y2)....]
        :param pixel_size: Pixel size
        :param tie_point: Tie point of the raster or tile
        :return: list of pixel locations
        """
        if type(coords_list) != list:
            coords_list = [coords_list]

        return list(((coord[0] - tie_point[0]) // pixel_size[0],
                     (coord[1] - tie_point[1]) // pixel_size[1])
                    if coord is not None else [None, None]
                    for coord in coords_list)

    def get_bounds(self,
                   xy_coordinates=True,
                   use_nodatavalue=False,
                   return_datasource=False,
                   return_memory_dataset=False,
                   with_buffer=0.0):
        """
        Method to return a list of raster coordinates
        :param xy_coordinates: return a list of xy coordinates if true,
                              else return [xmin, xmax, ymin, ymax]
                              Ignored if use_nodatavalue is True
        :param use_nodatavalue: If the nodatavalue should be used to vectorize areas with data
                                This will return a wkt geometry
        :param return_datasource: if bounds data sorce object should be returned
        :param return_memory_dataset: If a dataset in memory should be returned (/vsimem object)
        :param with_buffer: the geometry is buffered by specified pixels if > 0
        :return:
              if use_nodatavalue flag is False and xy_coordinates is False: list of numbers
              if use_nodatavalue is False and xy_coordinates is True: list of lists

              if use_nodatavalue is True: wkt_geometry
              if return_memory_dataset is selected: returns /vsimem/{basename}_bounds.shp
        """
        if not self.init:
            self.initialize()
        tie_pt = [self.transform[0], self.transform[3]]

        if not use_nodatavalue:
            if xy_coordinates:
                return [tie_pt,
                        [tie_pt[0] + self.metadict['xpixel'] * self.shape[2], tie_pt[1]],
                        [tie_pt[0] + self.metadict['xpixel'] * self.shape[2],
                         tie_pt[1] - self.metadict['ypixel'] * self.shape[1]],
                        [tie_pt[0], tie_pt[1] - self.metadict['ypixel'] * self.shape[1]],
                        tie_pt]
            else:
                return [tie_pt[0], tie_pt[0] + self.metadict['xpixel'] * self.shape[2],
                        tie_pt[1] - self.metadict['ypixel'] * self.shape[1], tie_pt[1]]
        else:

            mask_ras = self.nodata_mask(in_memory=True)
            basename = Handler(self.name).basename.split('.')[0]
            out_name = None

            if return_memory_dataset:
                out_name = '/vsimem/{}_bounds.shp'.format(basename)
                out_driver = ogr.GetDriverByName('ESRI Shapefile')
                out_datasource = out_driver.CreateDataSource('/vsimem/{}_bounds.shp'.format(basename))

            else:
                out_driver = ogr.GetDriverByName('Memory')
                out_datasource = out_driver.CreateDataSource('mem_source')

            out_spref = osr.SpatialReference()
            out_spref.ImportFromWkt(self.crs_string)

            out_layer = out_datasource.CreateLayer('{}_bounds'.format(basename),
                                                   srs=out_spref,
                                                   geom_type=OGR_TYPE_DEF['multipolygon'])

            out_field = ogr.FieldDefn('tmp_field', OGR_FIELD_DEF['int'])
            out_layer.CreateField(out_field)

            gdal.Polygonize(mask_ras.datasource.GetRasterBand(1),
                            None,
                            out_layer,
                            0)

            out_datasource.FlushCache()

            geom_wkt = None
            del_feat_id = []
            nfeat = out_layer.GetFeatureCount()

            for feat_indx in range(nfeat):
                feat = out_layer.GetFeature(feat_indx)
                items = feat.items()
                if items['tmp_field'] != 1:
                    out_layer.DeleteFeature(feat.GetFID())
                else:
                    geom = feat.GetGeometryRef()
                    geom.CloseRings()
                    if with_buffer > 0:
                        geom = geom.Buffer(self.transform[0]*float(with_buffer))

                    geom_wkt = geom.ExportToWkt()

            out_datasource.ExecuteSQL('REPACK ' + out_layer.GetName())
            out_datasource.ExecuteSQL('RECOMPUTE EXTENT ON ' + out_layer.GetName())

            if return_datasource:
                return out_datasource

            elif return_memory_dataset:
                out_datasource = None
                return out_name

            else:

                mask_ras = None
                out_datasource = None
                return geom_wkt

    def centroid(self,
                 use_nodatavalue=False):
        """
        Method to return centroid of the Raster object based on its boundary
        :param use_nodatavalue: Boolean flag to calculate the centroid
                                based on valid pixel boundary
        """
        bounds = self.get_bounds(use_nodatavalue=use_nodatavalue,
                                 xy_coordinates=False)

        if use_nodatavalue:
            bounds_geom = ogr.CreateGeometryFromWkt(bounds)
            centroid_geom = bounds_geom.Centroid()
            return centroid_geom.GetPoint()[0:2]

        else:
            bounds = self.get_bounds()
            return float(bounds[0] + bounds[1])/2.0, float(bounds[2] + bounds[3])/2.0

    def get_pixel_bounds(self,
                         bound_coords=None,
                         coords_type='pixel'):
        """
        Method to return image bounds in the format xmin, xmax, ymin, ymax
        :param bound_coords: (xmin, xmax, ymin, ymax)
        :param coords_type: type of coordinates specified in bound_coords: 'pixel' for pixel coordinates,
                                                                           'crs' for image reference system coordinates
        :return: tuple: (xmin, xmax, ymin, ymax) in pixel coordinates
        """
        if not self.init:
            self.initialize()

        if bound_coords is not None:
            if coords_type == 'pixel':
                xmin, xmax, ymin, ymax = bound_coords
            elif coords_type == 'crs':
                _xmin, _xmax, _ymin, _ymax = bound_coords
                coords_list = [(_xmin, _ymax), (_xmax, _ymax), (_xmax, _ymin), (_xmin, _ymin)]
                coords_locations = np.array(self.get_locations(coords_list,
                                                               (self.transform[1], self.transform[5]),
                                                               (self.transform[0], self.transform[3])))
                xmin, xmax, ymin, ymax = \
                    int(coords_locations[:, 0].min()), \
                    int(coords_locations[:, 0].max()), \
                    int(coords_locations[:, 1].min()), \
                    int(coords_locations[:, 1].max())
            else:
                raise ImageProcessingError("Unknown coordinate types")

            if xmin < 0:
                xmin = 0
            if xmax >= self.shape[2]:
                xmax = self.shape[2] - 1
            if ymin < 0:
                ymin = 0
            if ymax >= self.shape[1]:
                ymax = self.shape[1] - 1

            if xmin >= xmax:
                raise ImageProcessingError("Image x-size should be greater than 0")
            if ymin >= ymax:
                raise ImageProcessingError("Image y-size should be greater than 0")
        else:
            xmin, xmax, ymin, ymax = 0, self.shape[2] - 1, 0, self.shape[1] - 1

        return xmin, xmax, ymin, ymax

    def make_tile_grid(self,
                       tile_xsize=1024,
                       tile_ysize=1024,
                       bound_coords=None,
                       coords_type='pixel',
                       tile_buffer=None):
        """
        Returns the coordinates of the blocks to be extracted
        :param tile_xsize: Number of columns in the tile block
        :param tile_ysize: Number of rows in the tile block
        :param bound_coords: (xmin, xmax, ymin, ymax)
        :param coords_type: type of coordinates specified in bound_coords: 'pixel' for pixel coordinates,
                                                                           'crs' for image reference system coordinates
        :param tile_buffer: Buffer outside the tile boundary in image projection units
        :return: list of lists
        """
        if not self.init:
            self.initialize()

        self.tile_grid = []

        # convert to the number of pixels in the buffer region
        if tile_buffer is not None:
            buf_size_x = np.ceil(float(tile_buffer)/abs(float(self.transform[1])))
            buf_size_y = np.ceil(float(tile_buffer)/abs(float(self.transform[5])))
        else:
            buf_size_x, buf_size_y = None, None

        xmin, xmax, ymin, ymax = self.get_pixel_bounds(bound_coords,
                                                       coords_type)

        for y in range(ymin, ymax, tile_ysize):

            if y + tile_ysize < ymax:
                rows = tile_ysize
            else:
                rows = ymax - y

            for x in range(xmin, xmax, tile_xsize):
                if x + tile_xsize < xmax:
                    cols = tile_xsize
                else:
                    cols = xmax - x

                tie_pt = self.get_coords([(x, y)],
                                         (self.transform[1], self.transform[5]),
                                         (self.transform[0], self.transform[3]),
                                         pixel_center=False)[0]

                bounds = [tie_pt,
                          [tie_pt[0] + self.transform[1] * cols, tie_pt[1]],
                          [tie_pt[0] + self.transform[1] * cols, tie_pt[1] + self.transform[5] * rows],
                          [tie_pt[0], tie_pt[1] + self.transform[5] * rows],
                          tie_pt]

                self.tile_grid.append({'block_coords': (x, y, cols, rows),
                                       'tie_point': tie_pt,
                                       'bound_coords': bounds,
                                       'first_pixel': (xmin, ymin),
                                       'edge_buffer': (buf_size_x, buf_size_y)})

        self.ntiles = len(self.tile_grid)

    def get_tile(self,
                 bands=None,
                 block_coords=None,
                 finite_only=True,
                 edge_buffer=None,
                 nan_replacement=None,
                 n_tries=5,
                 nodatavalue=-9999):
        """
        Method to get raster numpy array of a tile
        :param bands: bands to get in the array, index starts from 0. (default: all)
        :param finite_only:  If only finite values should be returned
        :param edge_buffer: Number of extra pixels to retrieve further out from the edges (default: None)
                           Tuple (buf_size_x, buf_size_y)
        :param nan_replacement: replacement for NAN values
        :param n_tries: Number of randomized tries when reading raster is errored out
        :param nodatavalue: No data value default -9999
        :param block_coords: coordinates of tile to retrieve in image/array coords
                             format is (upperleft_x, upperleft_y, tile_cols, tile_rows)
                             upperleft_x and upperleft_y are array coordinates starting at 0,
                             cols and rows are number of pixels to retrieve for the tile along x and y respectively
        :return: numpy array
        """

        if edge_buffer is None:
            edge_buffer_x, edge_buffer_y = 0, 0
        else:
            edge_buffer_x, edge_buffer_y = edge_buffer

        if not self.init:
            self.initialize()

        if nan_replacement is None:
            if self.nodatavalue is None:
                nan_replacement = 0
            else:
                nan_replacement = self.nodatavalue

        if bands is None:
            bands = list(range(0, self.shape[0]))

        if block_coords is None:
            raise ImageProcessingError("Block coords needed to retrieve tile")
        else:
            upperleft_x, upperleft_y, tile_cols, tile_rows = block_coords

        # raster shape param
        ras_rows, ras_cols = self.shape[1], self.shape[2]

        # accounting for number of pixels less than the required size (always >= 0)
        # pixel deficit on left, top, right, and bottom edges respectively
        pixel_deficit = [(edge_buffer_x - upperleft_x) if (upperleft_x < edge_buffer_x) else 0,  # left

                         (edge_buffer_y - upperleft_y) if (upperleft_y < edge_buffer_y) else 0,  # top

                         (ras_cols - upperleft_x - tile_cols + 1) if
                         (ras_cols - upperleft_x - tile_cols + 1) < edge_buffer_x else 0,  # right

                         (ras_rows - upperleft_y - tile_rows + 1) if
                         (ras_rows - upperleft_y - tile_rows + 1) < edge_buffer_y else 0]  # bottom

        # corners
        new_upperleft_x = (upperleft_x - edge_buffer_x) + pixel_deficit[0]
        new_upperleft_y = (upperleft_y - edge_buffer_y) + pixel_deficit[1]

        # new block coordinates: xoffset, yoffset, xsize, ysize
        # per https://gdal.org/python/osgeo.gdal.Dataset-class.html#ReadAsArray
        new_block_coords = [new_upperleft_x,
                            new_upperleft_y,
                            tile_cols + (2 * edge_buffer_x - pixel_deficit[0] + pixel_deficit[2]),
                            tile_rows + (2 * edge_buffer_y - pixel_deficit[1] + pixel_deficit[3])]

        if self.array is None:

            tile_arr = np.zeros((len(bands),
                                 new_block_coords[3],
                                 new_block_coords[2]),
                                gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype))

            for jj, band in enumerate(bands):
                read_success = True
                for i_try in range(n_tries):
                    try:
                        temp_band = self.datasource.GetRasterBand(band + 1)
                        tile_arr[jj, :, :] = temp_band.ReadAsArray(*new_block_coords)
                        break
                    except RuntimeError as e:
                        read_success = False
                        Opt.cprint(e)
                        Opt.cprint('Retrying failed read...attempt {}'.format(i_try + 1))
                        time.sleep(random.random() + 1.0)

                if not read_success:
                    Opt.cprint('Retrying failed with {} attempts...using no-data value'.format(str(n_tries)))
                    if self.nodatavalue is not None:
                        tile_arr[jj, :, :] = self.nodatavalue
                    else:
                        tile_arr[jj, :, :] = nodatavalue

            if finite_only:
                if np.isnan(tile_arr).any() or np.isinf(tile_arr).any():
                    tile_arr[np.isnan(tile_arr)] = nan_replacement
                    tile_arr[np.isinf(tile_arr)] = nan_replacement

        else:
            bands = np.array([bands])
            tile_arr = self.array[bands,
                                  new_block_coords[1]:(new_block_coords[1] + new_block_coords[3]),
                                  new_block_coords[0]:(new_block_coords[0] + new_block_coords[2])]

        return tile_arr

    def get_next_tile(self,
                      tile_xsize=1024,
                      tile_ysize=1024,
                      bands=None,
                      get_array=True,
                      finite_only=True,
                      edge_buffer=None,
                      nan_replacement=None):

        """
        Generator to extract raster tile by tile
        :param tile_xsize: Number of columns in the tile block
        :param tile_ysize: Number of rows in the tile block
        :param bands: List of bands to extract (default: None, gets all bands; Index starts at 0)
        :param get_array: If raster array should be retrieved as well
        :param finite_only: If only finite values should be returned
        :param edge_buffer: Tuple of extra pixels to retrieve further out from the edges (default: None)
                            (edge_buf_x, edge_buf_y)
        :param nan_replacement: replacement for NAN values
        :return: Yields tuple: (tiepoint xy tuple, tile numpy array(2d array if only one band, else 3d array)
        """

        if not self.init:
            self.initialize()

        if self.ntiles is None:
            self.make_tile_grid(tile_xsize,
                                tile_ysize)
        if nan_replacement is None:
            if self.nodatavalue is None:
                nan_replacement = 0
            else:
                nan_replacement = self.nodatavalue

        if bands is None:
            bands = range(0, int(self.shape[0]))
        elif type(bands) in (int, float):
            bands = [int(bands)]
        elif type(bands) in (list, tuple):
            if all(list(isinstance(elem, str) for elem in bands)):
                bands = [self.bnames.index(elem) for elem in bands]
        else:
            raise ValueError('Unknown/unsupported data type for "bands" keyword')

        tile_counter = 0
        while tile_counter < self.ntiles:
            if get_array:
                tile_arr = self.get_tile(bands=bands,
                                         block_coords=self.tile_grid[tile_counter]['block_coords'],
                                         finite_only=finite_only,
                                         edge_buffer=edge_buffer,
                                         nan_replacement=nan_replacement)
            else:
                tile_arr = None

            yield self.tile_grid[tile_counter]['tie_point'], tile_arr

            tile_counter += 1

    def make_tiles(self,
                   tile_xsize=1024,
                   tile_ysize=1024,
                   out_path=None,
                   edge_buffer=None,
                   **kwargs):

        """
        Make and write smaller tile tif files from the raster tif file
        :param tile_xsize: Tile size along x
        :param tile_ysize: tile size along y
        :param edge_buffer: Buffer along the edges to include while making tiles
                            Tuple of (edge_buf_x, edge_buf_y) in number of pixels
        :param out_path: Output folder for the tiles
        :returns: None
        """
        if out_path is None:
            out_path = Handler(self.name).dirname

        tile_counter = 0
        for tie_pt, tile_arr in self.get_next_tile(tile_xsize=tile_xsize,
                                                   tile_ysize=tile_ysize,
                                                   edge_buffer=edge_buffer):

            block_coords = self.tile_grid[tile_counter]['block_coords']
            out_file = out_path + Handler(Handler(self.name).basename)\
                .add_to_filename('_{}'.format('_'.join([str(elem) for elem in block_coords])))

            out_transform = [tie_pt[0],
                             self.transform[1],
                             self.transform[2],
                             tie_pt[1],
                             self.transform[4],
                             self.transform[5]]

            tile_ras = Raster(out_file,
                              array=tile_arr,
                              bnames=self.bnames,
                              dtype=self.dtype,
                              shape=tile_arr.shape,
                              transform=out_transform,
                              crs_string=self.crs_string,
                              nodatavalue=self.nodatavalue)

            tile_ras.write_to_file(**kwargs)
            tile_ras = None
            tile_counter += 1

    @staticmethod
    def index_geom_wkt(wkt_strings,
                       geom_ids=None,
                       separate_multi_geom=True):

        """
        Method to index list of OGR geometry strings and
        split OGR multi-geometry wkt strings into single geometries

        :param wkt_strings: List of OGR geometry wkt strings
        :param geom_ids: List of geometry ids corresponding to wkt_string list
        :param separate_multi_geom: (bool) Flag to separate multigeometries into smaller geometries
        :returns: List of tuples [(geom_id: OGR geometry obj), ]
        """

        id_geom_list = []

        for wkt_string_indx, wkt_string in enumerate(wkt_strings):
            # multi geometry should be separated
            if separate_multi_geom:
                if ('MULTI' in wkt_string) or ('multi' in wkt_string):

                    # if multi geometry should be separated then add M prefix to index and
                    # add another index of the geometry after underscore
                    multi_geom = ogr.CreateGeometryFromWkt(wkt_string)

                    for multi_geom_indx in range(multi_geom.GetGeometryCount()):
                        geom_internal_id = '{}_{}'.format(str(wkt_string_indx), str(multi_geom_indx))\
                            if geom_ids is None else '{}_{}'.format(str(geom_ids[wkt_string_indx]),
                                                                    str(multi_geom_indx))
                        id_geom_list.append((geom_internal_id,
                                             multi_geom.GetGeometryRef(multi_geom_indx)))

                else:
                    # if no multi geometry in the string
                    id_geom_list.append((str(wkt_string_indx) if geom_ids is None else geom_ids[wkt_string_indx],
                                         ogr.CreateGeometryFromWkt(wkt_string)))

            else:
                # if multi geometry should not be separated
                id_geom_list.append((str(wkt_string_indx) if geom_ids is None else geom_ids[wkt_string_indx],
                                     ogr.CreateGeometryFromWkt(wkt_string)))

        return id_geom_list

    def extract_geom(self,
                     wkt_strings,
                     geom_ids=None,
                     band_order=None,
                     **kwargs):
        """
        Extract all pixels that intersect a geometry or a list of geometries in a Raster.
        The raster object should be initialized before using this method.
        Currently this method only supports single geometries per query.
        :param wkt_strings: List or Tuple of Vector geometries (e.g. point) in WKT string format
                           this geometry should be in the same CRS as the raster
                           Currently only 'Point' or 'MultiPoint' geometry is supported.
                           Accepted wkt_strings: List of POINT or MULTIPOINT wkt(s)
        :param geom_ids: List of geometry IDs
                        If for a MultiGeom only one ID is presented,
                        it will be suffixed with the order of the geometry part
        :param band_order: Order of bands to be extracted (list of band indices)  default: all bands

        :param kwargs: List of additional arguments
                        tile_size : tuple (x size, y size)
                                   Size of internal tiling
                        multi_geom_separate: bool (False) default
                                            if multi geometries should be separated or not
                        pass_pixel_coords: bool (False) default
                                           if coordinates of the pixels should be passed along the output
                        pixel_center: bool (True) default
                                     Used only if the pass_pixel_coords flag is set to true.
                                     Returns the pixel centers of each pixel if set as true
                                     else returns top left corners
                        min_pixels: Minimum number of pixels for reducer to calculate a value.
                                    If the number of pixels are less than this, then an empty list is returned
                        replace: If the pixel values in Raster.array object should be replaced with
                                 the reduced value? This will only be computed of min_pixels condition is
                                 satisfied and will change the Raster array. Use with caution.
                        reducer: Valid keywords: 'mean','median','max',
                                                 'min', 'percentile_xx' where xx is percentile from 1-99

        :return: Dictionary of dictionaries
                {internal_id: {'values': [[band1, ], ], 'coordinates': [(x1, y1), ]}, }
                if geom_id is supplied then internal_id is the supplied geom_id

        todo: Implement separation of overlapping feature extractions

        """
        if band_order is None:
            band_order = list(range(self.shape[0]))

        # define tile size
        if 'tile_size' in kwargs:
            tile_size = kwargs['tile_size']
        else:
            tile_size = (self.shape[1], self.shape[2])

        # if multi geometries should be separated or not
        if 'separate_multi_geom' in kwargs:
            separate_multi_geom = kwargs['separate_multi_geom']
        else:
            separate_multi_geom = False

        if 'pass_pixel_coords' in kwargs:
            pass_pixel_coords = kwargs['pass_pixel_coords']
        else:
            pass_pixel_coords = False

        if 'pixel_center' in kwargs:
            pixel_center = kwargs['pixel_center']
        else:
            pixel_center = True

        if 'reducer' in kwargs:
            reducer = kwargs['reducer']
        else:
            reducer = None

        if 'min_pixels' in kwargs:
            min_pixels = kwargs['min_pixels']
        else:
            min_pixels = None

        if 'replace' in kwargs:
            replace = kwargs['replace']
        else:
            replace = False

        # define band order
        if band_order is None:
            band_order = list(range(0, self.shape[0]))

        # initialize raster
        if not self.init or self.array is None:
            self.initialize()

        # make wkt strings into a list
        if type(wkt_strings) not in (list, tuple):
            wkt_strings = [wkt_strings]

        # list of geometry indices and OGR SWIG geometry objects
        # each dict entry contains   geom_id : geom
        id_geom_list = self.index_geom_wkt(wkt_strings,
                                           geom_ids=geom_ids,
                                           separate_multi_geom=separate_multi_geom)

        # make internal tiles
        self.make_tile_grid(*tile_size)

        # prepare dict struct
        out_geom_extract = dict()
        for internal_id, _ in id_geom_list:
            out_geom_extract[internal_id] = {'values': [], 'coordinates': [], 'xyz_loc': []}

        geom_id_order = list(internal_id for internal_id, _ in id_geom_list)

        # list of sample ids
        for tile in self.tile_grid:

            # create tile geometry from bounds
            tile_geom = ogr.CreateGeometryFromWkt('POLYGON(({}))'.format(', '.join(list(' '.join([str(x), str(y)])
                                                                                        for (x, y)
                                                                                        in tile['bound_coords']))))

            tile_arr = self.get_tile(block_coords=tile['block_coords'])

            # check if the geometry intersects and
            # place all same geometry types together
            geom_by_type = {}
            for samp_id, samp_geom in id_geom_list:
                if tile_geom.Intersects(samp_geom):
                    geom_type = samp_geom.GetGeometryType()
                    if geom_type not in geom_by_type:
                        geom_by_type[geom_type] = [(samp_id, samp_geom)]
                    else:
                        geom_by_type[geom_type].append((samp_id, samp_geom))

            # check is any geoms are available
            if len(geom_by_type) > 0:
                for geom_type, geom_list in geom_by_type.items():

                    # get tile shape and tie point
                    _, _, rows, cols = tile['block_coords']
                    tie_pt_x, tie_pt_y = tile['tie_point']

                    # create tile empty raster in memory
                    target_ds = gdal.GetDriverByName('MEM').Create('tmp',
                                                                   cols,
                                                                   rows,
                                                                   1,
                                                                   gdal.GDT_UInt16)

                    # set pixel size and tie point
                    target_ds.SetGeoTransform((tie_pt_x,
                                               self.transform[1],
                                               0,
                                               tie_pt_y,
                                               0,
                                               self.transform[5]))

                    # set raster projection
                    target_ds.SetProjection(self.crs_string)

                    # create vector in memory
                    burn_driver = ogr.GetDriverByName('Memory')
                    burn_datasource = burn_driver.CreateDataSource('mem_source')
                    burn_spref = osr.SpatialReference()
                    burn_spref.ImportFromWkt(self.crs_string)
                    burn_layer = burn_datasource.CreateLayer('tmp_lyr',
                                                             srs=burn_spref,
                                                             geom_type=geom_type)

                    # attributes
                    fielddefn = ogr.FieldDefn('fid', ogr.OFTInteger)
                    result = burn_layer.CreateField(fielddefn)
                    layerdef = burn_layer.GetLayerDefn()

                    geom_burn_val = 1
                    geom_dict = {}
                    for geom_id, geom in geom_list:
                        # create features in layer
                        temp_feature = ogr.Feature(layerdef)
                        temp_feature.SetGeometry(geom)
                        temp_feature.SetField('fid', geom_burn_val)
                        burn_layer.CreateFeature(temp_feature)
                        geom_dict[geom_burn_val] = geom_id
                        geom_burn_val += 1

                    gdal.RasterizeLayer(target_ds,
                                        [1],
                                        burn_layer,
                                        None,  # transformer
                                        None,  # transform
                                        [1],
                                        ['ALL_TOUCHED=TRUE',
                                         'ATTRIBUTE=FID'])

                    # read mask band as array
                    temp_band = target_ds.GetRasterBand(1)
                    mask_arr = temp_band.ReadAsArray()

                    min_burn_val = min(list(geom_dict.keys()))
                    max_burn_val = max(list(geom_dict.keys()))

                    pixel_xy_loc = np.where((mask_arr >= min_burn_val) & (mask_arr <= max_burn_val))
                    burn_vals = mask_arr[pixel_xy_loc]
                    pixel_xy_loc_list = np.array(list(zip(*pixel_xy_loc)))

                    burn_dict = {}
                    for geom_burn_val in range(min_burn_val, max_burn_val + 1):
                        burn_dict[geom_burn_val] = pixel_xy_loc_list[np.where(burn_vals == geom_burn_val)].tolist()

                    for geom_burn_val, geom_id in geom_dict.items():
                        # make list of unmasked pixels
                        if pass_pixel_coords:
                            # get coordinates
                            out_geom_extract[geom_id]['coordinates'] += self.get_coords(burn_dict[geom_burn_val],
                                                                                        (self.transform[1],
                                                                                         self.transform[5]),
                                                                                        tile['tie_point'],
                                                                                        pixel_center)

                        # get band values from tile array
                        out_geom_extract[geom_id]['values'] += list(tile_arr[band_order, x, y].tolist()
                                                                    for x, y in burn_dict[geom_burn_val])

                        out_geom_extract[geom_id]['xyz_loc'] += list([[band_index] + list(xy_loc)]
                                                                     for xy_loc in burn_dict[geom_burn_val]
                                                                     for band_index in band_order)

            if reducer is not None:
                warned = False
                for geom_id, geom_dict in out_geom_extract.items():
                    try:
                        n_pixels = len(geom_dict['values'])
                        if min_pixels is None:
                            min_pixels = n_pixels

                        if n_pixels >= min_pixels:
                            geom_dict['values'] = Sublist.reduce(geom_dict['values'],
                                                                 method=reducer,
                                                                 axis=0).tolist()

                            if replace:
                                pixel_loc = (np.array(geom_dict['xyz_loc'])[:, 0],
                                             np.array(geom_dict['xyz_loc'])[:, 1],
                                             np.array(geom_dict['xyz_loc'])[:, 2])

                                self.array[pixel_loc] = geom_dict['values']

                        else:
                            geom_dict['values'] = []

                        geom_dict['coordinates'] = Sublist.reduce(geom_dict['coordinates'],
                                                                  method=reducer,
                                                                  axis=0).tolist()
                    except Exception as e:
                        if not warned:
                            warnings.warn(e.args[0] + '\nNot reducing.'.format(reducer))
                            warned = True

                    out_geom_extract[geom_id] = geom_dict

        return list(out_geom_extract[internal_id] for internal_id in geom_id_order)

    def clip_by_geom(self,
                     cutline_file=None,
                     cutline_blend=0,
                     outfile=None,
                     return_vrt=False,
                     return_vrt_dict=False,
                     cutline_as_mask=True,
                     **creation_options):
        """
        Method to clip a raster to a given geometry/vector
        This method only supports clipping to the first layer in the datasource
        :param cutline_file: Shapefile, etc. to clip raster
        :param cutline_blend: blend distance in pixels
        :param outfile: Output filename
        :param return_vrt: If a VRT object should be returned instead of raster
        :param return_vrt_dict: if a VRT options dictionary should be returned instead
        :param cutline_as_mask: Use cutline extent for output bounds (default: true)
        :param creation_options: Other creation options to input
        :return: Raster object

        valid warp options can be found at:
        https://gdal.org/python/osgeo.gdal-module.html#WarpOptions:
        """
        cutline_ds = ogr.Open(cutline_file)
        layer_count = cutline_ds.GetLayerCount()
        cutline_layer = cutline_ds.GetLayer(0)
        cutline_layer_name = cutline_layer.GetDescription()

        if layer_count > 1:
            warnings.warn('Using top layer {} as cutline, ignoring others'.format(cutline_layer_name))

        cutline_ds = cutline_layer = None

        vrt_dict = dict()

        vrt_dict['cutlineDSName'] = cutline_file
        vrt_dict['cutlineLayer'] = cutline_layer_name
        vrt_dict['cutlineBlend'] = cutline_blend
        vrt_dict['cropToCutline'] = cutline_as_mask
        vrt_dict['copyMetadata'] = True

        creation_options_list = []
        if len(creation_options) > 0:
            for key, value in creation_options.items():
                creation_options_list.append('{}={}'.format(key.upper(),
                                                            value.upper()))
            vrt_dict['creationOptions'] = creation_options_list

        vrt_opt = gdal.WarpOptions(**vrt_dict)

        if return_vrt_dict:
            return vrt_dict
        else:
            if outfile is None:
                outfile = Handler(self.name).add_to_filename('_clipped')

            vrt_ds = gdal.Warp(outfile, self.name, options=vrt_opt)

            if return_vrt:
                return vrt_ds
            else:
                vrt_ds = None
                return Raster(outfile)

    def clip_by_extent(self,
                       xmin,
                       ymin,
                       xmax,
                       ymax,
                       outfile=None,
                       verbose=False,
                       **kwargs):
        """
        Method to clip a raster by its CRS extent - xmin, ymin, xmax, ymax
        :param xmin: x coordinate of lower left corner
        :param ymin: y coordinate of lower left corner
        :param xmax: x coordinate of upper right corner
        :param ymax y coordinate of upper right corner
        :param outfile: output file name (geotiff). If 'None' then the output will be written to memory
                        and this method will return a GDAL datasource object
        :param verbose: If some of the steps should be displayed
        :param kwargs: Keyword arguments (ignored if outfile is None):
                            examples: compress='lzw', bigtiff='yes'
        """

        if not self.init:
            self.initialize()

        pxmin, pxmax, pymin, pymax = self.get_pixel_bounds(bound_coords=(xmin, xmax, ymin, ymax),
                                                           coords_type='crs')

        ncols = pxmax - pxmin + 1
        nrows = pymax - pymin + 1

        if verbose:
            Opt.cprint('pxmin {}, pymin {}, pxmax {}, pymax {}, ncols {}, nrows {}'.format(str(pxmin),
                                                                                           str(pymin),
                                                                                           str(pymax),
                                                                                           str(pymax),
                                                                                           str(ncols),
                                                                                           str(nrows)))

        tile_arr = self.get_tile(block_coords=(pxmin, pymin, ncols, nrows))
        transform = (xmin,
                     self.transform[1],
                     self.transform[2],
                     ymax,
                     self.transform[4],
                     self.transform[5])

        outras = Raster(outfile,
                        array=tile_arr,
                        bnames=self.bnames,
                        dtype=self.dtype,
                        shape=tile_arr.shape,
                        transform=transform,
                        crs_string=self.crs_string,
                        nodatavalue=self.nodatavalue)

        if verbose:
            Opt.cprint(outras)

        if outfile is None:
            return outras.write_to_file(driver='MEM')

        else:
            outras.write_to_file(**kwargs)

            if verbose:
                Opt.cprint('Written: {}'.format(outfile))
            return

    def reproject(self,
                  outfile=None,
                  out_epsg=None,
                  out_wkt=None,
                  out_proj4=None,
                  out_spref=None,
                  output_res=None,
                  out_datatype=gdal.GDT_Float32,
                  resampling=None,
                  output_bounds=None,
                  out_format='GTiff',
                  out_nodatavalue=None,
                  verbose=False,
                  cutline_file=None,
                  cutline_blend=0,
                  return_datasource=False,
                  **creation_options):
        """
        Method to reproject raster object
        :param outfile: output file name
        :param out_epsg: EPSG code for output spatial reference
        :param out_wkt: WKT representation of output spatial reference
        :param out_proj4: PROJ4 representation of output spatial reference
        :param out_spref: Output spatial reference object
        :param output_res: output spatial resolution (xRes, yRes)
        :param out_datatype: output type (gdal.GDT_Byte, etc...)
        :param resampling: near, bilinear, cubic, cubicspline,
                           lanczos, average, mode, max, min, med, q1, q3
        :param out_nodatavalue: output no-data value to replace input no-data value
        :param output_bounds: output bounds as (minX, minY, maxX, maxY) in target SRS
        :param out_format: output format ("GTiff", etc...)
        :param verbose: If the steps should be displayed
        :param cutline_file: Shapefile, etc. to clip raster
        :param cutline_blend: blend distance in pixels
        :param return_datasource: If datasource object should be returned instead of raster
        :param creation_options: Creation options to be used while writing the raster
                                (example for geotiff: 'compress=lzw' , 'bigtiff=yes' )
        :return: VRT object or None (if output file is also specified)

        All the other valid warp options can be found at
        https://gdal.org/python/osgeo.gdal-module.html#WarpOptions
        """

        warp_dict = dict()

        if output_bounds is not None:
            warp_dict['outputBounds'] = output_bounds

        if output_res is not None:
            warp_dict['xRes'] = output_res[0]
            warp_dict['yRes'] = output_res[1]

        if out_nodatavalue is not None:
            warp_dict['dstNodata'] = out_nodatavalue
        else:
            warp_dict['dstNodata'] = self.nodatavalue

        warp_dict['srcNodata'] = self.nodatavalue

        if resampling is not None:
            warp_dict['resampleAlg'] = resampling
        else:
            warp_dict['resampleAlg'] = 'near'

        if verbose:
            Opt.cprint('Outfile: {}'.format(outfile))

        if out_spref is not None:
            spref = out_spref
        else:
            spref = osr.SpatialReference()

            if out_epsg is not None:
                res = spref.ImportFromEPSG(out_epsg)
            elif out_wkt is not None:
                res = spref.ImportFromWkt(out_wkt)
            elif out_proj4 is not None:
                res = spref.ImportFromProj4(out_proj4)
            else:
                res = spref.ImportFromWkt(self.crs_string)

        warp_dict['srcSRS'] = self.crs_string
        warp_dict['dstSRS'] = spref.ExportToWkt()

        warp_dict['outputType'] = out_datatype
        warp_dict['copyMetadata'] = True
        warp_dict['format'] = out_format

        if cutline_file is not None:
            clip_opts = self.clip_by_geom(cutline_file,
                                          cutline_blend,
                                          return_vrt_dict=True)

            warp_dict.update(clip_opts)

        creation_options_list = []
        if len(creation_options) > 0:
            for key, value in creation_options.items():
                creation_options_list.append('{}={}'.format(str(key).upper(),
                                                            str(value).upper()))

            warp_dict['creationOptions'] = creation_options_list

        warp_opts = gdal.WarpOptions(**warp_dict)

        if outfile is None:
            outfile = Handler(self.name).dirname + Handler().sep + '_reproject.tif'

        warp_datasource = gdal.Warp(outfile,
                                    self.name,
                                    options=warp_opts)

        if return_datasource:
            return warp_datasource
        else:
            vrt_ds = None


class MultiRaster:
    """
    Virtual raster class to manipulate GDAL virtual raster object
    """

    def __init__(self,
                 filelist=None,
                 initialize=True,
                 get_array=False):

        """
        Class constructor
        :param filelist: List of raster (.tif) files
        :param initialize: if the rasters from file list should be initialized
        :param get_array: if raster arrays should be read to memory
        """

        self.filelist = filelist
        self.rasters = list()

        if filelist is not None:

            if type(filelist).__name__ not in ('list', 'tuple'):
                filelist = [filelist]
            for filename in filelist:
                ras = Raster(filename)
                if initialize:
                    ras.initialize(get_array=get_array)
                self.rasters.append(ras)

        self.intersection = None
        self.nodatavalue = list(raster.nodatavalue for raster in self.rasters)
        self.resolutions = list((raster.transform[1], raster.transform[5]) for raster in self.rasters)

    def get_extent(self,
                   index=None,
                   intersection=True,
                   verbose=False,
                   _return=True):
        """
        Method to get intersecting bounds of the raster objects
        :param index: list of indices of raster files/objects
        :param verbose: If some of the steps should be displayed
        :param intersection: If intersection (True) should be returned or union (False) (Default: True)
        :param _return: Should the method return the bound coordinates
        :return: coordinates of intersection (minx, miny, maxx, maxy)
        """

        wkt_list = list()
        if index is not None:
            for ii in index:
                bounds = self.rasters[ii].get_bounds()
                wktstring = 'POLYGON(({}))'.format(', '.join(list(' '.join([str(x), str(y)])
                                                                  for (x, y) in bounds)))
                wkt_list.append(wktstring)
        else:
            for raster in self.rasters:
                bounds = raster.get_bounds()
                wktstring = 'POLYGON(({}))'.format(', '.join(list(' '.join([str(x), str(y)])
                                                                  for (x, y) in bounds)))
                wkt_list.append(wktstring)

        geoms = list(ogr.CreateGeometryFromWkt(wktstring) for wktstring in wkt_list)

        if intersection:
            temp_geom = geoms[0]
            for geom in geoms[1:]:
                temp_geom = temp_geom.Intersection(geom)
        else:
            temp_geom = ogr.Geometry(ogr.wkbMultiPolygon)

            for geom in geoms:
                temp_geom.AddGeometryDirectly(geom)

        temp_geom_json = temp_geom.ExportToJson()
        json_str = json.loads(temp_geom_json)

        temp_coords = []
        for level1_list in json_str['coordinates']:
            for level2_list in level1_list:
                if json_str['type'] == 'Polygon':
                    temp_coords.append(level2_list)
                elif json_str['type'] == 'MultiPolygon':
                    for level3_list in level2_list:
                        temp_coords.append(level3_list)
                else:
                    temp_coords.append(None)

        minx = min(list(coord[0] for coord in temp_coords))
        miny = min(list(coord[1] for coord in temp_coords))

        maxx = max(list(coord[0] for coord in temp_coords))
        maxy = max(list(coord[1] for coord in temp_coords))

        self.intersection = (minx, miny, maxx, maxy)

        if verbose:
            Opt.cprint('xmin: {}, ymin: {}, xmax: {}, ymax: {}'.format(*[str(elem)
                                                                         for elem in
                                                                         self.intersection]))

        if _return:
            return self.intersection

    def layerstack(self,
                   order=None,
                   bands=None,
                   verbose=False,
                   outfile=None,
                   return_vrt=False,
                   write_vrt=False,
                   extent='inclusive',
                   **kwargs):

        """
        Method to layerstack rasters in a given order
        :param order: order of raster layerstack
        :param verbose: If some of the steps should be printed to console
        :param outfile: Name of the output file (.tif)
        :param bands: list of band numbers (index start at 0)
        :param extent: 'inclusive' or 'exclusive' for all the files in the stack
        :param return_vrt: If the file should be written to disk or vrt object should be returned
        :param write_vrt: If the vrt file should be written to disk
        :return: None

        valid build vrt options in kwargs
        (from https://gdal.org/python/osgeo.gdal-module.html#BuildVRT):

        valid translate options in kwargs
        (from https://gdal.org/python/osgeo.gdal-module.html#TranslateOptions):
        """

        if order is None:
            order = list(range(len(self.rasters)))

        vrt_dict = dict()

        if bands is not None:
            vrt_dict['bandList'] = (np.array([bands]) + 1).tolist()
        else:
            vrt_dict['bandList'] = [1]

        if 'output_bounds' in kwargs:
            vrt_dict['outputBounds'] = kwargs['output_bounds']
        else:
            vrt_dict['outputBounds'] = self.intersection

        output_res = min(list(np.abs(self.resolutions[i][0]) for i in order))

        if 'outputresolution' in kwargs:
            vrt_dict['xRes'] = kwargs['outputresolution'][0]
            vrt_dict['yRes'] = kwargs['outputresolution'][1]
        else:
            vrt_dict['xRes'] = output_res
            vrt_dict['yRes'] = output_res

        if 'nodatavalue' in kwargs:
            vrt_dict['srcNodata'] = kwargs['nodatavalue']
        else:
            vrt_dict['srcNodata'] = self.nodatavalue[0]

        if 'outnodatavalue' in kwargs:
            vrt_dict['VRTNodata'] = kwargs['outnodatavalue']
        else:
            vrt_dict['VRTNodata'] = self.nodatavalue[0]

        if 'resample' in kwargs:
            vrt_dict['resampleAlg'] = kwargs['resample']
        else:
            vrt_dict['resampleAlg'] = 'cubic'

        if verbose:
            Opt.cprint('Getting bounds ...')

        intersection_dict = {'inclusive': False,
                             'exclusive': True}
        vrt_dict['outputBounds'] = self.get_extent(index=order,
                                                   intersection=intersection_dict[extent])
        vrt_dict['separate'] = True
        vrt_dict['hideNodata'] = False

        if verbose:
            Opt.cprint('Files: \n{}'.format('\n'.join(list(self.filelist[i] for i in order))))

        _vrt_opt_ = gdal.BuildVRTOptions(**vrt_dict)

        if outfile is None:
            return_vrt = True
            outfile = Handler(self.filelist[0]).dirname + Handler().sep + 'layerstack1.tif'

        if write_vrt:
            vrtfile = outfile.replace('.tif', '.vrt')
        else:
            vrtfile = '/vsimem/' + Opt.temp_name()

        _vrt_ = gdal.BuildVRT(vrtfile,
                              list(self.filelist[i] for i in order),
                              options=_vrt_opt_)

        if not return_vrt:
            trns_dict = {'creationOptions': []}

            if 'compress' in kwargs:
                trns_dict['creationOptions'].append('COMPRESS={}'.format(str(kwargs['compress']).upper()))

            if 'bigtiff' in kwargs:
                trns_dict['creationOptions'].append('BIGTIFF={}'.format(str(kwargs['bigtiff']).upper()))

            if verbose:
                Opt.cprint('Writing layer stack file : {} ...'.format(outfile))

            trns_opts = gdal.TranslateOptions(**trns_dict)

            gdal.Translate(outfile,
                           _vrt_,
                           options=trns_opts)
            _vrt_ = None

            if verbose:
                Opt.cprint('Done!')
        else:
            return _vrt_

    def composite(self,
                  layer_indices=None,
                  verbose=False,
                  outfile=None,
                  composite_type='mean',
                  tile_size=1024,
                  write_raster=False,
                  **kwargs):
        """
        Method to calculate raster composite in a given order
        :param layer_indices: list of layer indices
        :param verbose: If some of the steps should be printed to console
        :param outfile: Name of the output file (.tif)
        :param tile_size: Size of internal tile
        :param write_raster: If the raster file should be written or a raster object be returned
        :param composite_type: mean, median, pctl_xxx (eg: pctl_5, pctl_99, pctl_100, etc.),
        :return: None
        """

        if layer_indices is None:
            layer_indices = list(range(len(self.rasters)))  # list of layer indices
            t_order = list(range(1, len(self.rasters) + 1))  # list of bands to include in raster tiles
        else:
            t_order = list(elem + 1 for elem in layer_indices)  # list of bands to include in raster tiles

        # layer stack vrt
        _ls_vrt_ = self.layerstack(order=layer_indices,
                                   verbose=verbose,
                                   return_vrt=True,
                                   **kwargs)

        # raster object from vrt
        lras = Raster('tmp_layerstack')
        lras.datasource = _ls_vrt_
        lras.initialize()

        if 'bound_coords' in kwargs:
            if 'coords_type' in kwargs:
                lras.make_tile_grid(tile_size,
                                    tile_size,
                                    bound_coords=kwargs['bound_coords'],
                                    coords_type=kwargs['coords_type'])
            else:
                lras.make_tile_grid(tile_size,
                                    tile_size,
                                    bound_coords=kwargs['bound_coords'],
                                    coords_type='crs')
        else:
            lras.make_tile_grid(tile_size,
                                tile_size)

        if verbose:
            Opt.cprint(lras)

        # make numpy array to hold the final result
        out_arr = np.zeros((lras.shape[1], lras.shape[2]),
                           dtype=gdal_array.GDALTypeCodeToNumericTypeCode(lras.dtype))

        # loop through raster tiles
        count = 0
        for tie_pt, tile_arr in lras.get_next_tile(bands=t_order):

            if verbose:
                Opt.cprint(lras.tile_grid[count]['block_coords'])

            _x, _y, _cols, _rows = lras.tile_grid[count]['block_coords']

            if composite_type == 'mean':
                temp_arr = np.apply_along_axis(lambda x: np.mean(x[x != lras.nodatavalue]), 0, tile_arr)
            elif composite_type == 'median':
                temp_arr = np.apply_along_axis(lambda x: np.median(x[x != lras.nodatavalue]), 0, tile_arr)
            elif composite_type == 'min':
                temp_arr = np.apply_along_axis(lambda x: np.min(x[x != lras.nodatavalue]), 0, tile_arr)
            elif composite_type == 'max':
                temp_arr = np.apply_along_axis(lambda x: np.max(x[x != lras.nodatavalue]), 0, tile_arr)
            elif 'pctl' in composite_type:
                pctl = int(composite_type.split('_')[1])
                temp_arr = np.apply_along_axis(lambda x: np.percentile(x[x != lras.nodatavalue], pctl), 0, tile_arr)
            else:
                raise ImageProcessingError('Unknown composite option')

            # update output array with tile composite
            out_arr[_y: (_y+_rows), _x: (_x+_cols)] = temp_arr
            count += 1

        # write array to raster
        lras.array = out_arr

        if write_raster:
            # write raster
            lras.write_to_file(outfile)
            if verbose:
                Opt.cprint('Written {}'.format(outfile))
        else:
            return lras

    def mosaic(self,
               order=None,
               verbose=False,
               outfile=None,
               nodata_values=None,
               bands=None,
               blend_pixels=10,
               blend_cutline=None,
               add_overviews=False,
               return_datasource=False,
               **kwargs):
        """
        Method to mosaic rasters in a given order
        This method uses all bands of the given rasters
        :param order: order of raster layerstack
        :param verbose: If some of the steps should be printed to console
        :param outfile: Name of the output file (.tif)
        :param nodata_values: Value or tuple (or list) of values used as
                             nodata bands for each image to be mosaicked
        :param blend_pixels: width of pixels to blend around the cutline or
                     raster boundary for multiple rasters (default: 10)
        :param bands: List of band indices (index starts at 0)
        :param blend_cutline: vector file (shapefile) in memory or on disk
                              to use for blending
        :param add_overviews: If overviews should be added to the resulting image
        :param return_datasource: If datasource of the mosaic output should be returned
        :param kwargs: Other options
                       valid warp options in kwargs :
                       https://gdal.org/python/osgeo.gdal-module.html#WarpOptions
        :return: None
        todo: 1) band selection: Use VRT to get bands that will be used at runtime
              2) blending options: image bounds buffered at intersection with other image
              using:
                from scipy.ndimage.morphology import binary_erosion
                from scipy.spatial.distance import cdist
                def dist_from_edge(img):
                    I = binary_erosion(img) # Interior mask
                    C = img - I             # Contour mask
                    out = C.astype(int)     # Setup o/p and assign cityblock distances
                    out[I] = cdist(np.argwhere(C), np.argwhere(I), 'cityblock').min(0) + 1
                    return out
                    OR
                    import numpy as np
                    from scipy.ndimage import distance_transform_cdt
                    def distance_from_edge(x):
                        x = np.pad(x, 1, mode='constant')
                        dist = distance_transform_cdt(x, metric='taxicab')
                        return dist[1:-1, 1:-1]
                    OR
                    import numpy as np
                    from scipy.spatial.distance import cdist
                    def feature_dist(input):
                        # Takes a labeled array as returned by scipy.ndimage.label and
                        # returns an intra-feature distance matrix.
                        I, J = np.nonzero(input)
                        labels = input[I,J]
                        coords = np.column_stack((I,J))
                        sorter = np.argsort(labels)
                        labels = labels[sorter]
                        coords = coords[sorter]
                        sq_dists = cdist(coords, coords, 'sqeuclidean')
                        start_idx = np.flatnonzero(np.r_[1, np.diff(labels)])
                        nonzero_vs_feat = np.minimum.reduceat(sq_dists, start_idx, axis=1)
                        feat_vs_feat = np.minimum.reduceat(nonzero_vs_feat, start_idx, axis=0)
                        return np.sqrt(feat_vs_feat)
                        This approach requires O(N2) memory,
                        where N is the number of nonzero pixels.
                        If this is too demanding,
                        you could "de-vectorize" it along one axis (add a for-loop).
        """

        if order is None:
            order = list(range(len(self.filelist)))

        if verbose:
            Opt.cprint('Files: \n{}'.format('\n'.join(list(self.filelist[i] for i in order))))

        warp_dict = {'options': []}

        output_res = max(list(np.abs(self.resolutions[i][0]) for i in order))

        if 'output_resolution' in kwargs:
            warp_dict['xRes'] = kwargs['output_resolution'][0]
            warp_dict['yRes'] = kwargs['output_resolution'][1]
        else:
            warp_dict['xRes'] = output_res
            warp_dict['yRes'] = output_res

        warp_dict['srcSRS'] = self.rasters[0].crs_string

        if 'out_crs_string' in kwargs:
            warp_dict['dstSRS'] = kwargs['out_crs_string']
        else:
            warp_dict['dstSRS'] = warp_dict['srcSRS']

        if nodata_values is not None:
            warp_dict['srcNodata'] = nodata_values
        else:
            warp_dict['srcNodata'] = list(ras.nodatavalue for ras in self.rasters)

        if 'outnodatavalue' in kwargs:
            warp_dict['dstNodata'] = kwargs['outnodatavalue']
        else:
            warp_dict['dstNodata'] = self.nodatavalue[0]

        if 'resample' in kwargs:
            warp_dict['resampleAlg'] = kwargs['resample']
        else:
            warp_dict['resampleAlg'] = 'bilinear'

        if verbose:
            Opt.cprint('Getting bounds ...')

        warp_dict['outputBounds'] = self.get_extent(index=order,
                                                    verbose=verbose,
                                                    intersection=False)
        warp_dict['targetAlignedPixels'] = True
        warp_dict['outputType'] = self.rasters[0].dtype

        if blend_cutline is not None:
            blend_datasource = ogr.Open(blend_cutline)
            blend_layer = blend_datasource.GetLayerByIndex(0)
            blend_layer_name = blend_layer.GetName()

            warp_dict['cutlineDSName'] = blend_cutline
            warp_dict['cutlineLayer'] = blend_layer_name
            warp_dict['cutlineBlend'] = blend_pixels

        warp_dict['creationOptions'] = []

        if 'compress' in kwargs:
            warp_dict['creationOptions'].append('COMPRESS={}'.format(str(kwargs['compress']).upper()))

        if 'bigtiff' in kwargs:
            warp_dict['creationOptions'].append('BIGTIFF={}'.format(str(kwargs['bigtiff']).upper()))

        warp_opts = gdal.WarpOptions(**warp_dict)

        if bands is not None:
            filelist = list(ras.get_bands(bands,
                                          return_vrt=True,
                                          return_raster=False) for ras in self.rasters)
        else:
            filelist = self.filelist

        warp_datasource = gdal.Warp(outfile,
                                    filelist,
                                    options=warp_opts)

        if add_overviews:
            ras = Raster(outfile)
            if type(add_overviews) == list:
                ras.add_overviews(overviews=add_overviews)
            else:
                ras.add_overviews()

        if return_datasource:
            return warp_datasource

